"""This module contains unit tests for :mod:`~prody.proteins`."""

from numpy.testing import *
from prody.utilities import importDec
dec = importDec()

from prody import *
from prody import LOGGER
from prody.tests import unittest
from prody.tests.datafiles import *

import os

LOGGER.verbosity = 'none'

PATH_1UBI = pathDatafile('1ubi')
SPLIT_PATH = os.path.split(PATH_1UBI)
EXPECT_PATH = os.path.join(SPLIT_PATH[0], 'addH_' + SPLIT_PATH[1])

class TestFixer(unittest.TestCase):

    def setUp(self):
        """Set PDB file data and parse the PDB file."""

        self.ubi_orig = DATA_FILES['1ubi']
        self.ubi_addH = DATA_FILES['1ubi_addH']

    @dec.slow
    def testPDBFixer(self):

        if prody.PY3K:
            self.filename = addMissingAtoms(PATH_1UBI, method='pdbfixer')
            self.assertEqual(self.filename, EXPECT_PATH,
                            'addMissing atoms wrote the file in the wrong place')
            
            atoms = parsePDB(self.filename)
            self.assertEqual(len(atoms.hydrogen), DATA_FILES['1ubi_addH']['n_h'],
                            'addMissing atoms did not add the right number of hydrogens')



# -*- coding: utf-8 -*-
"""This module defines :class:`KDTree` class for dealing with atomic coordinate
sets and handling periodic boundary conditions."""

from numbers import Number
from numpy import array, ndarray, concatenate, empty

from prody import LOGGER

def createKDTreeByDim(KDTclass, coords, bucketsize):
    kdt = KDTclass(3, bucketsize)
    kdt.set_data(coords)
    return kdt

def createKDTreeByCoords(KDTclass, coords, bucketsize):
    kdt = KDTclass(coords, bucketsize)
    return kdt

try:
    from ._CKDTree import KDTree as _KDTree
    CKDTree = lambda coords, bz : createKDTreeByDim(_KDTree, coords, bz) 
except ImportError:
    try:
        from Bio.PDB.kdtrees import KDTree as _KDTree
        CKDTree = lambda coords, bz : createKDTreeByCoords(_KDTree, coords, bz) 
    except ImportError:
        try:
            from Bio.KDTree._CKDTree import KDTree as _KDTree
            CKDTree = lambda coords, bz : createKDTreeByDim(_KDTree, coords, bz) 
        except ImportError:
            raise ImportError('CKDTree module could not be imported. '
                            'Reinstall ProDy or install Biopython '
                            'to solve the problem.')

__all__ = ['KDTree']

_ = array([-1., 0., 1.])
REPLICATE = array([[x, y, z] for x in _ for y in _ for z in _])

class KDTree(object):

    """An interface to Thomas Hamelryck's C KDTree module that can handle
    periodic boundary conditions.  Both point and pair search are performed
    using the single :meth:`search` method and results are retrieved using
    :meth:`getIndices` and :meth:`getDistances`.

    **Periodic Boundary Conditions**

    *Point search*

    A point search around a *center*, indicated with a question mark (``?``)
    below, involves making images of the point in cells sharing a wall or an
    edge with the unitcell that contains the system.  The search is performed
    for all images of the *center* (27 in 3-dimensional space) and unique
    indices with the minimum distance from them to the *center* are returned.
    ::

          _____________________________
         |        1|        2|        3|
         |       ? |       ? |      ?  |
         |_________|_________|_________|
         |        4|o  h h  5|        6| ? and H interact in periodic image 4
         |       ?H| h  o  ? |      ?  | but not in the original unitcell (5)
         |_________|_________|_________|
         |        7|        8|        9|
         |       ? |       ? |      ?  |
         |_________|_________|_________|

    There are two requirements for this approach to work: (i) the *center* must
    be in the original unitcell, and (ii) the system must be in the original
    unitcell with parts in its immediate periodic images.

    *Pair search*

    A pair search involves making 26 (or 8 in 2-d) replicas of the system
    coordinates.  A KDTree is built for the system (``O`` and ``H``) and all
    its replicas (``o`` and ``h``).  After pair search is performed, unique
    pairs of indices and minimum distance between them are returned.
    ::

          _____________________________
         |o  h h  1|o  h h  2|o  h h  3|
        h| h  o   h| h  o   h| h  o    |
         |_________|_________|_________|
         |o  h h  4|O  H H  5|o  h h  6|
        h| h  o   H| H  O   h| h  o    |
         |_________|_________|_________|
         |o  h h  7|o  h h  8|o  h h  9|
        h| h  o   h| h  o   h| h  o    |
         |_________|_________|_________|

    Only requirement for this approach to work is that the system must be
    in the original unitcell with parts in its immediate periodic images.


    .. seealso::
       :func:`.wrapAtoms` can be used for wrapping atoms into the single
       periodic image of the system."""

    def __init__(self, coords, **kwargs):
        """
        :arg coords: coordinate array with shape ``(N, 3)``, where N is number
            of atoms
        :type coords: :class:`numpy.ndarray`, :class:`.Atomic`, :class:`.Frame`

        :arg unitcell: orthorhombic unitcell dimension array with shape
            ``(3,)``
        :type unitcell: :class:`numpy.ndarray`

        :arg bucketsize: number of points per tree node, default is 10
        :type bucketsize: int"""

        unitcell = kwargs.get('unitcell')
        if not isinstance(coords, ndarray):
            if unitcell is None:
                try:
                    unitcell = coords.getUnitcell()
                except AttributeError:
                    pass
                else:
                    if unitcell is not None:
                        LOGGER.info('Unitcell information from {0} will be '
                                    'used.'.format(str(coords)))
            try:
                # using getCoords() because coords will be stored internally
                # and reused when needed, this will avoid unexpected results
                # due to changes made to coordinates externally
                coords = coords.getCoords()
            except AttributeError:
                raise TypeError('coords must be a Numpy array or must have '
                                'getCoords attribute')
        else:
            coords = coords.copy()

        if coords.ndim != 2:
                raise Exception('coords.ndim must be 2')
        if coords.shape[-1] != 3:
                raise Exception('coords.shape must be (N,3)')
        if coords.min() <= -1e6 or coords.max() >= 1e6:
                raise Exception('coords must be between -1e6 and 1e6')

        self._bucketsize = kwargs.get('bucketsize', 10)

        if not isinstance(self._bucketsize, int):
            raise TypeError('bucketsize must be an integer')
        if self._bucketsize < 1:
            raise ValueError('bucketsize must be a positive integer')

        self._coords = None
        self._unitcell = None
        self._neighbors = None
        if unitcell is None:
            self._kdtree = CKDTree(coords, self._bucketsize)
        else:
            if not isinstance(unitcell, ndarray):
                raise TypeError('unitcell must be a Numpy array')
            if unitcell.shape != (3,):
                raise ValueError('unitcell.shape must be (3,)')
            self._kdtree = CKDTree(coords, self._bucketsize)
            self._coords = coords
            self._unitcell = unitcell
            self._replicate = REPLICATE * unitcell
            self._kdtree2 = None
            self._pbcdict = {}
            self._pbckeys = []
            self._n_atoms = coords.shape[0]
        self._none = kwargs.pop('none', lambda: None)
        try:
            self._none()
        except TypeError:
            raise TypeError('none argument must be callable')
        self._oncall = kwargs.pop('oncall', 'both')
        assert self._oncall in ('both', 'dist'), 'oncall must be both or dist'

    def __call__(self, radius, center=None):
        """Shorthand method for searching and retrieving results."""

        self.search(radius, center)
        if self._oncall == 'both':
            return self.getIndices(), self.getDistances()
        elif self._oncall == 'dist':
            return self.getDistances()

    def search(self, radius, center=None):
        """Search pairs within *radius* of each other or points within *radius*
        of *center*.

        :arg radius: distance (Å)
        :type radius: float

        :arg center: a point in Cartesian coordinate system
        :type center: :class:`numpy.ndarray`"""

        if not isinstance(radius, Number):
            raise TypeError('radius must be a number')
        if radius <= 0:
            raise TypeError('radius must be a positive number')

        if center is not None:
            if not isinstance(center, ndarray):
                raise TypeError('center must be a Numpy array instance')
            if center.shape != (3,):
                raise ValueError('center.shape must be (3,)')

            if self._unitcell is None:
                self._kdtree.search_center_radius(center, radius)
                self._neighbors = None

            else:
                kdtree = self._kdtree
                search = kdtree.search_center_radius
                get_radii = lambda : get_KDTree_radii(kdtree)
                get_indices = lambda : get_KDTree_indices(kdtree)
                get_count = kdtree.get_count

                _dict = {}
                _dict_get = _dict.get
                _dict_set = _dict.__setitem__
                for center in center + self._replicate:
                    search(center, radius)
                    if get_count():
                        [_dict_set(i, min(r, _dict_get(i, 1e6)))
                         for i, r in zip(get_indices(), get_radii())]
                self._pbcdict = _dict
                self._pdbkeys = list(_dict)

        else:
            if self._unitcell is None:
                self._neighbors = self._kdtree.neighbor_search(radius)
            else:
                kdtree = self._kdtree2
                if kdtree is None:
                    coords = self._coords
                    coords = concatenate([coords + rep
                                          for rep in self._replicate])
                    kdtree = CKDTree(coords, self._bucketsize)
                    self._kdtree2 = kdtree
                n_atoms = len(self._coords)
                _dict = {}
                neighbors = kdtree.neighbor_search(radius)
                if kdtree.neighbor_get_count():
                    _get = _dict.get
                    _set = _dict.__setitem__

                    for nb in neighbors:
                        i = nb.index1 % n_atoms
                        j = nb.index2 % n_atoms
                        if i < j:
                            _set((i, j), min(nb.radius, _get((i, j), 1e6)))
                        elif j < i:
                            _set((j, i), min(nb.radius, _get((j, i), 1e6)))
                self._pbcdict = _dict
                self._pdbkeys = list(_dict)


    def getIndices(self):
        """Returns array of indices for points or pairs, depending on the type
        of the most recent search."""

        if self.getCount():
            if self._unitcell is None:
                if self._neighbors is None:
                    return get_KDTree_indices(self._kdtree)
                else:
                    return array([(n.index1, n.index2)
                                  for n in self._neighbors], int)
            else:
                return array(self._pdbkeys)
        return self._none()

    def getDistances(self):
        """Returns array of distances."""

        if self.getCount():
            if self._unitcell is None:
                if self._neighbors is None:
                    return get_KDTree_radii(self._kdtree)
                else:
                    return array([n.radius for n in self._neighbors])
            else:
                _dict = self._pbcdict
                return array([_dict[i] for i in self._pdbkeys])
        return self._none()

    def getCount(self):
        """Returns number of points or pairs."""

        if self._unitcell is None:
            if self._neighbors is None:
                return self._kdtree.get_count()
            else:
                return self._kdtree.neighbor_get_count()
        else:
            return len(self._pbcdict)

def get_KDTree_indices(kdtree):
    indices = None
    try:
        indices = kdtree.get_indices()
    except:
        n = kdtree.get_count()
        if n:
            indices = empty(n, int)
            kdtree.get_indices(indices)
    return indices

def get_KDTree_radii(kdtree):
    radii = None
    try:
        radii = kdtree.get_radii()
    except:
        n = kdtree.get_count()
        if n:
            radii = empty(n, 'f')
            kdtree.get_radii(radii)
    return radii



# -*- coding: utf-8 -*-
"""This module defines functions for query CATH database."""

from os import getcwd, remove
from os.path import sep as pathsep
from os.path import isdir, isfile, join, split, splitext, normpath

import re
from numbers import Integral

from prody import LOGGER, SETTINGS, PY3K
from prody.proteins import parsePDB
from prody.utilities import makePath, gunzip, relpath, copyFile, openURL
from prody.utilities import sympath, indentElement, isPDB, isURL, isListLike

if PY3K:
    import urllib.parse as urllib
    import urllib.request as urllib2
else:
    import urllib
    import urllib2

import xml.etree.ElementTree as ET

__all__ = ['CATHDB', 'CATHElement']

CATH_URL = 'http://download.cathdb.info'
CATH_NAME_PATH = '/cath/releases/latest-release/cath-classification-data/cath-names.txt'
CATH_DOMAIN_PATH = '/cath/releases/latest-release/cath-classification-data/cath-domain-list.txt'
CATH_BOUNDARY_PATH = '/cath/releases/latest-release/cath-classification-data/cath-domain-boundaries-seqreschopping.txt'

isCATH = re.compile('^[0-9.]+$').match
# Note that if the chain id is blank, it will be represented by "0" in CATH domain ID.
isDomain = re.compile('^[A-Za-z0-9]{5}[0-9]{2}$').match

class CATHElement(ET.Element):
    """This class handles individual elements from the cath tree 
    including tree navigation and the getting of associated CATH domains, 
    PDB IDs and selection strings."""
    def __init__(self, tag, attrib={}, parent=None, **extra):
        super(CATHElement, self).__init__(tag, attrib, **extra)
        self._parent = parent

    def __getitem__(self, index):
        if isinstance(index, Integral):
            return super(CATHElement, self).__getitem__(index)
        else:
            return self.attrib[index]

    def __repr__(self):
        if self.isDomain():
            return '<CATHElement: {0}>'.format(self.id)
        return '<CATHElement: {0} ({1} members)>'.format(self.id, len(self))

    def isDomain(self):
        return isDomain(self.id)

    def getID(self):
        return self.tag

    def setID(self, value):
        self.tag = value
    
    id = property(getID, setID)

    def getchildren(self):
        try:
            children = super(CATHElement, self).getchildren()
        except AttributeError:
            children = list(self)
        collection = CATHCollection(children, self)
        return collection

    children = property(getchildren)

    def getparent(self):
        return self._parent

    def setparent(self, value):
        self._parent = value
    
    parent = property(getparent, setparent)

    def find(self, key):
        for child in self:
            if child.tag == key:
                return child
        return None

    def numDomains(self):
        n_leaves = 0
        if self.isDomain():
            n_leaves = 1
        else:
            for child in self:
                if child.isDomain():
                    n_leaves += 1
                else:
                    n_leaves += child.numDomains()
        return n_leaves

    def getDomains(self):
        leaves = []
        if self.isDomain():
            leaves.append(self)
        else:
            for child in self:
                if child.isDomain():
                    leaves.append(child)
                else:
                    leaves.extend(child.getDomains())
        return leaves

    def getDomainData(self, attr):
        leaves = self.getDomains()
        data = []

        for leaf in leaves:
            data.append(leaf[attr])
        return data

    def getPDBs(self, chain=True):
        leaves = self.getDomains()
        pdbs = []

        for leaf in leaves:
            pdb_id = leaf.attrib['pdb']
            if chain:
                pdb_id += leaf.attrib['chain']
            pdbs.append(pdb_id)
        
        return pdbs
    
    def getSelstrs(self):
        leaves = self.getDomains()
        data = []

        for leaf in leaves:
            rangestr = leaf.attrib['range']
            selstr = range2selstr(rangestr)
            data.append(selstr)
        return data

    def getCATH(self):
        try:
            cath = self.attrib['cath']
        except KeyError:
            if self.parent is not None:
                cath = self.parent.cath
            else:
                cath = None # this is the case for the root
        return cath

    cath = property(getCATH)

    def getName(self):
        try: 
            name = self.attrib['name'].lstrip(':')
            if name == '':
                name = 'Superfamily %s'%self.cath
        except KeyError:
            if self.parent is not None:
                name = self.parent.name
            else:
                name = None # this is the case for the root
        return name

    name = property(getName)


    def parsePDBs(self, **kwargs):
        """Load PDB into memory as :class:`.AtomGroup` instances using :func:`.parsePDB` and 
        perform selection based on residue ranges given by CATH."""
        
        pdbs = self.getPDBs(True)
        selstrs = self.getSelstrs()
        header = kwargs.get('header', False)
        model = kwargs.get('model', None)

        LOGGER.timeit('_cath_parsePDB')
        LOGGER.info('Parsing {0} PDB files...'.format(len(pdbs)))
        ret = parsePDB(*pdbs, **kwargs)

        if model != 0:
            headers = None
            if header:
                prots, headers = ret
            else:
                prots = ret

            if not isinstance(prots, list):
                prots = [prots]

                if header:
                    headers = [headers]
                    ret = (prots, headers)
                else:
                    ret = prots
                    
            LOGGER.info('Extracting domains...')
            for i in range(len(prots)):
                sel = prots[i].select(selstrs[i])
                prots[i] = sel
        LOGGER.report('CATH domains are parsed and extracted in %.2fs', '_cath_parsePDB')

        return ret


class CATHCollection(CATHElement):
    """"This class handles collections of elements from the CATH tree 
    including the listing of names and IDs."""
    def __init__(self, items, element=None):
        if element is not None:
            tag = element.tag
            attrib = element.attrib
        else:
            tag = 'cath'
            attrib = {}

        super(CATHCollection, self).__init__(tag=tag, attrib=attrib)

        if not isListLike(items):
            items = [items]

        parents = []
        for item in items:
            self.append(item)
            parents.append(item.parent)

        uniq_parents = set(parents)
        if len(uniq_parents) == 1:
            self._parent = parents[0]

    def __repr__(self):
        names = []
        n = 0
        for child in self:
            n += 1
            if n > 100:
                names.append('...')
                break
            else:
                names.append(repr(child))

        return '<CATHCollection:\n[%s]>'%'\n'.join(names)

    def isDomain(self):
        return False

    def getCATH(self):
        cath = []
        for child in self:
            cath.append(child.attrib['cath'])
        return cath

    cath = property(getCATH)

    def getName(self):
        names = []
        for child in self:
            names.append(child.name)
        return names

    name = property(getName)

    def getID(self):
        tags = []
        for child in self:
            tags.append(child.tag)
        return tags
    
    id = property(getID)


class CATHDB(ET.ElementTree):
    """This class is for handing the data in the CATH database. 
    It facilitates tree navigation and ID-based searching."""

    def __init__(self, source=CATH_URL):
        """The class is initialised using data from *source*, which 
        is downloaded from the CATH website by default, but can 
        also be parsed from an XML file."""
        self.root = CATHElement('root')
        super(CATHDB, self).__init__(self.root)
        self._source = source
        self._raw_names = None
        self._raw_domains = None
        self._raw_boundaries = None
        self._map = {'root': self.root}

        # this property is only for debug purposes and will not be copied
        self._unclassified = {}

        if source is not None:
            self.update()
    
    def __repr__(self):
        return "<CATHDB: {0} members>".format(len(self.root))

    def reset(self):
        # by removing the immediate children of root, 
        # all the descendants will be removed
        for child in self.root:
            self.root.remove(child)

        self._unclassified = {}

    def update(self, source=None):
        """Update data and files from CATH."""

        self._source = source = self._source or source
        self.reset()
        if source is None:
            return

        LOGGER.timeit('_cath_update')
        
        type_ = 0
        tree = None
        if isinstance(source, str):
            if isfile(source):
                type_ = 1
            elif isURL(source):
                type_ = 0
            else:
                type_ = 2
        elif hasattr(source, 'read'):
                type_ = 1
        else:
            raise TypeError('source must be either an url, file name, file handle, '
                            'or text in xml format')

        if type_ == 0:
            LOGGER.info('Fetching data from CATH...')
            self._fetch()

            LOGGER.info('Parsing CATH files...')
            self._parse()
        elif type_ == 1:
            LOGGER.info('Reading data from the local xml file...')
            tree = ET.parse(source)
        elif type_ == 2:
            LOGGER.info('Parsing input string...')
            tree = ET.fromstring(source)

        # post-processing
        if type_ > 0:
            root = tree.getroot()
            nodes = root.iter()

            # remove prefix from node tags
            for node in nodes:
                node.tag = node.tag.lstrip('id.')

            # convert int to str
            length_nodes = root.findall('.//*[@length]')
            for node in length_nodes:
                node.attrib['length'] = int(node.attrib['length'])
            
            copy2(root, self.root)
            self._update_map()

        LOGGER.report('CATH local database built in %.2fs.', '_cath_update')

    def _update_map(self, parent=None):
        if parent is None:
            parent = self.root
        for child in parent:
            self._map[child.id] = child
            self._update_map(child)

    def copy(self):
        db = CATHDB(source=None)
        copy2(self.root, db.root)

        db._map = self._map
        db._source = self._source
        db._raw_names = self._raw_names
        db._raw_domains = self._raw_domains
        db._raw_boundaries = self._raw_boundaries
        return db

    def save(self, filename='cath.xml'):
        """Write local CATH database to an XML file. *filename* can either be a 
        file name or a handle."""

        LOGGER.timeit('_cath_write')

        if not isinstance(filename, str):
            try:
                fn = filename.name
            except AttributeError:
                fn = repr(filename)
            f = filename
        else:
            fn = filename

        LOGGER.info('Writing data to {0}...'.format(fn))

        if not len(self.root):
            raise ValueError('local database has not been built, '
                             'please call update() first')

        tree = self.copy()
        root = tree.getroot()

        # convert int to str
        length_nodes = root.findall('.//*[@length]')
        for node in length_nodes:
            node.attrib['length'] = str(node.attrib['length'])
        
        # add prefix to node tags
        nodes = root.iter()
        for node in nodes:
            node.tag = 'id.' + node.tag

        # add indentation to nodes
        indentElement(root)
        
        if isinstance(filename, str):
            f = open(filename, 'wb')
        tree.write(f, encoding='utf-8')
        f.close()

        LOGGER.report('CATH local database saved in %.2fs.', '_cath_write')

    def find(self, identifier='root'):
        """Find an :class:`xml.etree.Element` object given *identifier*."""

        if len(self.root) is None:
            raise ValueError('local database has not been built, '
                             'please call update() first')
        if identifier != 'root':
            try:
                data = self._map[identifier]
            except KeyError:
                iter = self.iter(identifier)
                try:
                    data = next(iter)
                except StopIteration:
                    data = None
        else:
            data = self.root
        return data

    def _fetch(self):
        """Download CATH files via HTTP."""

        try:
            response = urllib2.urlopen(CATH_URL + CATH_NAME_PATH)
            name_data = response.read()
        except Exception as error:
            raise type(error)('HTTP connection problem when extracting CATH names: '\
                              + repr(error))

        try:
            response = urllib2.urlopen(CATH_URL + CATH_DOMAIN_PATH)
            domain_data = response.read()
        except Exception as error:
            raise type(error)('HTTP connection problem when extracting domain information: '\
                              + repr(error))

        try:
            response = urllib2.urlopen(CATH_URL + CATH_BOUNDARY_PATH)
            boundary_data = response.read()
        except Exception as error:
            raise type(error)('HTTP connection problem when extracting domain boundaries: '\
                              + repr(error))

        #if PY3K:
        name_data = name_data.decode('utf-8')
        domain_data = domain_data.decode('utf-8')
        boundary_data = boundary_data.decode('utf-8')

        self._raw_names = name_data
        self._raw_domains = domain_data
        self._raw_boundaries = boundary_data
        
    def _parse(self):
        """Parse CATH files into a :class:`~xml.etree.ElementTree` structure. 
        The leaf nodes are tagged by domain names, e.g. 3hgnA02, and other nodes 
        are tagged by their full CATH identifiers, e.g. 2.40.10.10. The parsing 
        process assumes the domain file is organized in a topdown fashion, i.e. 
        1 -> 1.20 -> 1.20.120 -> 1.20.120.1000. """
        
        name_data = self._raw_names 
        domain_data = self._raw_domains
        boundary_data = self._raw_boundaries

        root = self.root

        # parsing the name file
        lines = name_data.splitlines()
        for line in lines:
            text = line.strip()
            if not text:
                continue
            if text.startswith('#'):
                continue
            items = text.split('    ')

            cath_id, domain_id, cath_name = items
            #pdb_id = domain_id[:4]
            #chain = domain_id[4]

            level = getLevel(cath_id)
            attrib = {'cath': cath_id,
                      'rep': domain_id,
                      'name': cath_name,}
                      #'pdb': pdb_id,
                      #'chain': chain}
            
            if level == 1:
                node = CATHElement(cath_id, attrib=attrib, parent=root)
                root.append(node)
                self._map[cath_id] = node
            else:
                parent_id = getParent(cath_id)
                #parent = findNodeByCATH(root, parent_id)
                parent = self.find(parent_id)
                if parent is None:
                    LOGGER.warn('error encountered when building the tree: {0}'
                                .format(cath_id))
                    continue
                node = CATHElement(cath_id, attrib=attrib, parent=parent)
                parent.append(node)
                self._map[cath_id] = node

        # parsing the domain file
        lines = domain_data.splitlines()
        for line in lines:
            # Column 1:  CATH domain name (seven characters)
            # Column 2:  Class number
            # Column 3:  Architecture number
            # Column 4:  Topology number
            # Column 5:  Homologous superfamily number
            # Column 6:  S35 sequence cluster number
            # Column 7:  S60 sequence cluster number
            # Column 8:  S95 sequence cluster number
            # Column 9:  S100 sequence cluster number
            # Column 10: S100 sequence count number
            # Column 11: Domain length
            # Column 12: Structure resolution (Angstroms)
            #            (999.000 for NMR structures and 1000.000 for obsolete PDB entries)
            text = line.strip()
            if not text:
                continue
            if text.startswith('#'):
                continue
            items = text.split()

            domain_id = items[0]
            pdb_id = domain_id[:4]
            chain = domain_id[4]
            if chain == '0':
                chain = ''
            
            cath_id = '.'.join(items[1:5])
            #S35, S60, S95, S100 = items[5:9]
            length = int(items[10])
            #resolution = float(items[11])

            attrib = {'cath': cath_id,
                      'pdb': pdb_id,
                      'chain': chain,
                      'length': length,
                      'range': ''}

            #parent = findNodeByCATH(root, cath_id)
            parent = self.find(cath_id)
            if parent is None:
                LOGGER.warn('error encountered when assigning domains: {0}'
                            .format(domain_id))
                continue
            node = CATHElement(domain_id, attrib=attrib, parent=parent)
            parent.append(node)
            self._map[domain_id] = node

        # parsing the boundary file
        lines = boundary_data.splitlines()
        for line in lines:
            # Example: 3hgnA02	13-111,228-240
            text = line.strip()
            if not text:
                continue
            if text.startswith('#'):
                continue
            items = text.split()

            domain_id, resrange = items

            if domain_id in self._map:
                node = self.find(domain_id)
            else:
                self._unclassified[domain_id] = resrange
                continue

            if node is None:
                LOGGER.warn('error encountered when assigning domain boundaries: {0}'
                            .format(domain_id))
                continue
            node.attrib['range'] = resrange
    
    def search(self, identifier):
        """Search the *identifier* against the local CATH database and return a list 
        of :class:`~xml.etree.ElementTree` instances. *identifier* can be a 4-digit 
        PDB ID plus an 1-digit chain ID (optional), 7-digit domain ID, or CATH ID."""

        ret = None
        if isCATH(identifier):
            ret = self.find(identifier)
        elif isDomain(identifier):
            identifier = identifier[:4].lower() + identifier[-3:]
            ret = self.find(identifier)
        elif isPDB(identifier):
            pdb_id = identifier[:4].lower()
            chain = identifier[4:]
            for gap in ' -_:':
                chain = chain.replace(gap, '')
            
            root = self.find()
            nodes = root.findall('.//*[@pdb="%s"]'%pdb_id)
            if chain.strip() != '':
                for i in range(len(nodes)-1, -1, -1):
                    node = nodes[i]
                    if node.attrib['chain'] != chain:
                        del nodes[i]
            ret = nodes
        else:
            raise ValueError('identifier can be a 4-digit PDB ID plus 1-digit chain ID '
                             '(optional), 7-digit domain ID, or CATH ID.')
        
        if ret is not None:
            if isinstance(ret, list):
                ret = CATHCollection(ret)
        return ret

# Note: all the following functions should be kept hidden from users
def splitCATH(cath_id, cumul=False):
    fields = cath_id.split('.')

    if cumul:
        fields_ = list(fields)
        for i in range(len(fields)):
            fields_[i] = '.'.join(fields[:i+1])
        fields = fields_
    return fields

def findNodeByCATH(root, cath_id):
    fields = splitCATH(cath_id, cumul=True)

    node = root
    for field in fields:
        node = node.find(field)
        if node is None:
            return None
    return node

def getParent(cath_id):
    fields = splitCATH(cath_id)
    parent = '.'.join(f for f in fields[:-1])
    return parent

def getLevel(cath_id):
    return len(splitCATH(cath_id))

def isLeaf(cath_id):
    level = getLevel(cath_id)

    if level > 4:
        raise ValueError('invalid cath_id: {0}'.format(cath_id))

    return level == 4

def copy2(a, b):
    """Copy a tree structure from *a* to *b*"""
    for c in a:
        node = CATHElement(c.tag, c.attrib, b)
        b.append(node)
        copy2(c, node)

def range2selstr(rangestr):
    if rangestr.strip() == '':
        return None
    frags = rangestr.split(',')
    sels = []
    for frag in frags:
        try:
            fromtos = frag.split('-')
            if len(fromtos) == 2:
                fro, to = fromtos
            else:
                LOGGER.warn('range "%s" is irregular'%rangestr)
                fro = '1'
                to = fromtos[-1]
            fro_num = intResnum(fro)
            to_num = intResnum(to)

            if fro_num > to_num:
                LOGGER.warn('range "%s" is irregular'%rangestr)
                to_num = fro_num
                fro_num = 1
            fro = str(fro_num)
            to = str(to_num)
        except ValueError:
            print('error occurred when parsing "%s"'%rangestr)
            continue
        sels.append('resindex %s to %s'%(fro, to))
    selstr = ' or '.join(sels)
    return selstr

def intResnum(resnum):
    match = re.findall('[0-9\\-]+', resnum)
    num = None
    if match:
        num = int(match[0])
    return num
    


"""This module contains unit tests for :mod:`prody.measure.transform` module.
"""

from numpy import zeros, ones, eye, all
from numpy.testing import assert_equal

from prody.tests import unittest
from prody.tests.datafiles import parseDatafile

from prody.measure import moveAtoms, wrapAtoms

UBI = parseDatafile('1ubi')


class TestMoveAtoms(unittest.TestCase):

    def testToArgument(self):

        atoms = UBI.ca

        coords = UBI.getCoords()
        center = atoms._getCoords().mean(0)
        moveAtoms(atoms, to=zeros(3), ag=True)
        assert_equal(UBI._getCoords(), coords - center)

    def testByArgument(self):

        atoms = UBI
        offset = ones(3) * 10.
        coords = atoms.getCoords()
        moveAtoms(atoms, by=offset)
        assert_equal(atoms._getCoords(), coords + offset)

    def testTransformation(self):

        atoms = UBI.ca
        coords = UBI.getCoords()
        matrix = eye(4)
        moveAtoms(atoms, by=matrix, ag=True)
        assert_equal(UBI._getCoords(), coords)


class TestWrapAtoms(unittest.TestCase):

    def testWrap(self):

        xyz = UBI.getCoords()
        xyzmin = xyz.min(0) - 0.5
        unitcell = xyz.max(0) - xyzmin + 0.5
        center = xyzmin - unitcell / 2
        wrapAtoms(UBI, unitcell, center)
        diff = xyz - UBI.getCoords()
        self.assertTrue(all(diff == unitcell))




__author__ = 'Ahmet Bakan, Anindita Dutta'

from prody.tests import TestCase
import os
from os.path import join

from numpy.testing import assert_array_equal

from prody.tests.datafiles import *
from prody.tests import TEMPDIR
from prody import MSAFile, parseMSA, LOGGER, writeMSA
from prody.utilities import createStringIO, importDec
dec = importDec()

LOGGER.verbosity = None

FASTA = parseMSA(pathDatafile('msa_Cys_knot.fasta'))
SELEX = parseMSA(pathDatafile('msa_Cys_knot.slx'))
STOCK = parseMSA(pathDatafile('msa_Cys_knot.sth'))
FASTA_LIST = list(MSAFile(pathDatafile('msa_Cys_knot.fasta')))
SELEX_LIST = list(MSAFile(pathDatafile('msa_Cys_knot.sth')))
STOCK_LIST = list(MSAFile(pathDatafile('msa_Cys_knot.sth')))

class TestMSAFile(TestCase):

    def testMSAFile(self):

        self.assertListEqual(FASTA_LIST, SELEX_LIST)
        self.assertListEqual(FASTA_LIST, STOCK_LIST)

    def testWriteFasta(self):

        fasta = createStringIO()
        with MSAFile(fasta, 'w', format='fasta') as out:
            for seq in MSAFile(pathDatafile('msa_Cys_knot.fasta')):
                out.write(seq)
        fasta.seek(0)
        fasta_list = list(MSAFile(fasta, format='fasta'))
        self.assertListEqual(FASTA_LIST, fasta_list)

    def testWriteSelex(self):

        selex = createStringIO()
        with MSAFile(selex, 'w', format='selex') as out:
            for seq in MSAFile(pathDatafile('msa_Cys_knot.slx')):
                out.write(seq)
        selex.seek(0)
        selex_list = list(MSAFile(selex, format='selex'))
        self.assertListEqual(SELEX_LIST, selex_list)

    def testWriteStockholm(self):

        stock = createStringIO()
        with MSAFile(stock, 'w', format='stockholm') as out:
            for seq in MSAFile(pathDatafile('msa_Cys_knot.sth'), split=False):
                out.write(seq)
        stock.seek(0)
        stock_list = list(MSAFile(stock, format='stockholm'))
        self.assertListEqual(STOCK_LIST, stock_list)

class TestParseMSA(TestCase):

    def testArray(self):

        assert_array_equal(FASTA._getArray(), SELEX._getArray())
        assert_array_equal(SELEX._getArray(), STOCK._getArray())

    def testIterator(self):

        self.assertListEqual(list(FASTA), list(SELEX))
        self.assertListEqual(list(FASTA), list(STOCK))

    def testMapping(self):

        self.assertDictEqual(FASTA._mapping, SELEX._mapping)
        self.assertDictEqual(FASTA._mapping, STOCK._mapping)

class TestWriteMSA(TestCase):

    def testSelex(self):
        filename = writeMSA(join(TEMPDIR, 'test.slx'), SELEX)
        selex = parseMSA(pathDatafile(filename))
        self.assertListEqual(list(SELEX), list(selex))
        if os.path.isfile(filename):
            os.remove(filename)

    def testFasta(self):
        filename = writeMSA(join(TEMPDIR, 'test.fasta.gz'), FASTA)
        fasta = list(MSAFile(pathDatafile(filename)))
        self.assertListEqual(list(FASTA), list(fasta))
        if os.path.isfile(filename):
            os.remove(filename)



# -*- coding: utf-8 -*-
"""This module contains features for accessing databases containing protein
related data.

Pfam
====

The following functions can be used to search and retrieve Pfam_ data:

  * :func:`.fetchPfamMSA` - download MSA files
  * :func:`.searchPfam` - search for domain families of a protein

.. _Pfam: https://www.ebi.ac.uk/interpro/entry/pfam/

UniProt
========

The following functions and class can be used to search and retrieve UniProt_ data:

  * :func:`.queryUniprot` - query UniProt and parse the results as a dictionary
  * :class:`.UniprotRecord` - a wrapper from UniProt data with functions including parsing PDBs
  * :func:`.searchUniprot` - search UniProt and return a UniprotRecord


.. _UniProt: https://www.uniprot.org/

CATH
====

The following class and its functions can be used to search and retrieve CATH_ data:

  * :class:`.CATHDB` - parse, handle and navigate the tree-like structure of the CATH database

.. _CATH: http://download.cathdb.info

DALI
====

The following class and functions can be used to search and retrieve data using the DALI_ structure alignment server:

  * :func:`.searchDali` - search for similar structures using DALI
  * :class:`.DaliRecord` - fetch and handle outputs from DALI searches
  * :func:`.daliFilterMultimers` - filter DALI results to obtain multimers of a particular size

.. _DALI: http://ekhidna2.biocenter.helsinki.fi/dali/

QuartataWeb
============

The following classes and functions can be used to search and retrieve data using the QuartataWeb_ structure alignment server:

  * :class:`.QuartataWebBrowser` - class based on the Splinter web browser package to search QuartataWeb
  * :class:`.QuartataChemicalRecord` - class to handle the outputs of QuartataWeb searches
  * :func:`.searchQuartataWeb` - perform QuartataWeb searches and return the output in a QuartataChemicalRecord

.. _QuartataWeb: http://quartata.csb.pitt.edu

Gene Ontology Annotation (GOA)
================================

The following classes and functions can be used to search and retrieve data from the EBI GOA_ database:

  * :func:`.queryGOA` - query GOA using a PDB ID
  * :class:`.GOADictList` - class to handle data from GOA queries
  * :func:`.parseOBO` - parse an OBO file containing the Gene Ontology.
  * :func:`.parseGAF` - parse a Gene Association File (GAF)
  * :func:`.showGoLineage` - visualize GO tree
  * :func:`.calcGoOverlap` - Calculate overlap between GO terms from their distance in the graph

.. _GOA: https://www.ebi.ac.uk/GOA/

Interpro
====

The following functions can be used to search and retrieve Interpro_ data:

  * :func:`.searchInterpro` - search for domain families of a protein

.. _Interpro: https://www.ebi.ac.uk/interpro/

BioExcel-CV19
====

The following functions can be used to retrieve BioExcel-CV19_ data:

  * :func:`.fetchBioexcelPDB` - fetch PDB files for starting structures for trajectories

.. _BioExcel-CV19: https://bioexcel-cv19-dev.bsc.es/

"""

__all__ = []

from . import pfam
from .pfam import *
__all__.extend(pfam.__all__)

from . import uniprot
from .uniprot import *
__all__.extend(uniprot.__all__)

from . import cath
from .cath import *
__all__.extend(cath.__all__)

from . import dali
from .dali import *
__all__.extend(dali.__all__)

from . import goa
from .goa import *
__all__.extend(goa.__all__)

from . import quartataweb
from .quartataweb import *
__all__.extend(quartataweb.__all__)

from . import interpro
from .interpro import *
__all__.extend(interpro.__all__)

from . import bioexcel
from .bioexcel import *
__all__.extend(bioexcel.__all__)



"""This module defines some sequence evolution applications."""

import importlib
import sys

try:
    import argparse
except ImportError:
    from .. import argparse

from prody.apps.apptools import *
from prody.utilities.misctools import impLoadModule

if sys.version_info[0] == 2:
    import imp
    path_prody = imp.find_module('prody')[1]
else:
    path_prody = importlib.util.find_spec("prody").submodule_search_locations[0]

try:
    import imp
    path_apps = imp.find_module('apps', [path_prody])[1]
    path_apps = imp.find_module('evol_apps', [path_apps])[1]
except ModuleNotFoundError:
    path_apps = importlib.util.find_spec("prody.apps").submodule_search_locations[0]
    path_apps += '/evol_apps/'

EVOL_APPS = ['search', 'fetch', 'filter', 'refine', 'merge', 'occupancy',
             'conserv', 'coevol', 'rankorder']

__all__ = ['evol_main']

evol_parser = argparse.ArgumentParser(
    description="Evol: Sequence Evolution and Dynamics Analysis",
    epilog="See 'evol <command> -h' for more information on a specific "
           "command."
    )

evol_parser.add_argument('-c', '--cite',
    help="print citation info and exit",
    action=ProDyCitation, nargs=0)

evol_parser.add_argument('-v', '--version',
    help="print ProDy version and exit",
    action=ProDyVersion, nargs=0)

evol_parser.add_argument('-e', '--examples', action=UsageExample, nargs=0,
    help='show usage examples and exit')


evol_parser.set_defaults(usage_example=
"""Sequence coevolution analysis involves several steps that including
retrieving data and refining it for calculations.  These steps are
illustrated below for RnaseA protein family.

Search Pfam database:

  $  evol search 2w5i

Download Pfam MSA file:

  $  evol fetch RnaseA

Refine MSA file:

  $ evol refine RnaseA_full.slx -l RNAS1_BOVIN --seqid 0.98 --rowocc 0.8

Checking occupancy:

  $ evol occupancy RnaseA_full.slx -l RNAS1_BOVIN -o col -S

Conservation analysis:

  $ evol conserv RnaseA_full_refined.slx

Coevolution analysis:

  $ evol coevol RnaseA_full_refined.slx -S -c apc

Rank order analysis (you can use a PDB file if you write it first):

  $ evol rankorder RnaseA_full_refined_mutinfo_corr_apc.txt -p \
2W5IB_3-121.pdb --seq-sep 3 """, test_examples=[(0,1,2)])

evol_commands = evol_parser.add_subparsers(
    title='subcommands')


for cmd in EVOL_APPS:
    cmd = 'evol_' + cmd
    mod = impLoadModule('prody.apps.evol_apps.', cmd, path_apps)
    mod.APP.addApplication(evol_commands)


def evol_main():

    if len(sys.argv) == 1:
        evol_parser.print_help()
    else:
        namespace = evol_parser.parse_args()
        namespace.func(namespace)


if __name__ == '__main__':
    evol_main()



"""This module defines utility functions for drawing."""
# Note for developers: this module should be imported only a specific function is needed. 
# Please do not import this module when initializing ProDy for robustness reasons.

import numpy as np

from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
from matplotlib import ticker

__all__ = ['Arrow3D', 'drawArrow3D', 'drawTree', 'IndexFormatter']

class Arrow3D(FancyArrowPatch):
    """This function is implemented by tacaswell on stackoverflow: 
    https://stackoverflow.com/a/29188796."""
    def __init__(self, xs, ys, zs, *args, **kwargs):
        super(Arrow3D, self).__init__((0,0), (0,0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
        FancyArrowPatch.draw(self, renderer)

def drawArrow3D(*xyz, **kwargs):
    from matplotlib.pyplot import gca

    allargs = xyz
    xs = []; ys = []; zs = []; args = []

    for arg in allargs:
        if not np.isscalar(arg):
            if len(arg) == 3:
                xs.append(arg[0])
                ys.append(arg[1])
                zs.append(arg[2])
                continue
        args.append(arg)

    ax = gca()
    if 'arrowstyle' not in kwargs:
        kwargs['arrowstyle'] = '-|>'
    if 'mutation_scale' not in kwargs:
        kwargs['mutation_scale'] = 20
    if 'color' not in kwargs:
        kwargs['color'] = 'k'

    arrow = Arrow3D(xs, ys, zs, *args, **kwargs)
    ax.add_artist(arrow)
    return arrow

def drawTree(
    tree,
    label_func=str,
    show_confidence=False,
    # For power users
    orientation='horizontal',
    inverted=False,
    branch_labels=None,
    label_colors=None,
    *args,
    **kwargs
):
    """Plot the given tree using matplotlib. This function is adapted 
    from :func:`~Bio.Phylo._utils.draw` for more versatile usages.
    """
    import matplotlib.pyplot as plt
    import matplotlib.collections as mpcollections

    # Arrays that store lines for the plot of clades
    horizontal_linecollections = []
    vertical_linecollections = []

    orientation = orientation.lower()
    if orientation == 'v':
        orientation = 'vertical'
    elif orientation == 'h':
        orientation = 'horizontal'

    linewidth = kwargs.pop('linewidth', 1.)
    linewidth = kwargs.pop('lw', linewidth)

    fontsize = kwargs.pop('fontsize', 10)

    # Options for displaying branch labels / confidence
    def conf2str(conf):
        if int(conf) == conf:
            return str(int(conf))
        return str(conf)

    if not branch_labels:
        if show_confidence:

            def format_branch_label(clade):
                try:
                    confidences = clade.confidences
                    # phyloXML supports multiple confidences
                except AttributeError:
                    pass
                else:
                    return "/".join(conf2str(cnf.value) for cnf in confidences)
                if clade.confidence is not None:
                    return conf2str(clade.confidence)
                return None

        else:

            def format_branch_label(clade):
                return None

    elif isinstance(branch_labels, dict):

        def format_branch_label(clade):
            return branch_labels.get(clade)

    else:
        if not callable(branch_labels):
            raise TypeError(
                "branch_labels must be either a dict or a callable (function)"
            )
        format_branch_label = branch_labels

    # options for displaying label colors.
    if label_colors:
        if callable(label_colors):

            def get_label_color(label):
                return label_colors(label)

        else:
            # label_colors is presumed to be a dict
            def get_label_color(label):
                return label_colors.get(label, "black")

    else:

        def get_label_color(label):
            # if label_colors is not specified, use black
            return "black"

    # Layout

    def get_x_positions(tree):
        """Create a mapping of each clade to its horizontal position.

        Dict of {clade: x-coord}
        """
        depths = tree.depths()
        # If there are no branch lengths, assume unit branch lengths
        if not max(depths.values()):
            depths = tree.depths(unit_branch_lengths=True)

        if inverted:
            max_depth = max(depths.values())
            for clade in depths:
                depths[clade] = max_depth - depths[clade]
        return depths

    def get_y_positions(tree):
        """Create a mapping of each clade to its vertical position.

        Dict of {clade: y-coord}.
        Coordinates are negative, and integers for tips.
        """
        maxheight = tree.count_terminals()
        # Rows are defined by the tips
        heights = {
            tip: maxheight - i for i, tip in enumerate(reversed(tree.get_terminals()))
        }

        # Internal nodes: place at midpoint of children
        def calc_row(clade):
            for subclade in clade:
                if subclade not in heights:
                    calc_row(subclade)
            # Closure over heights
            heights[clade] = (
                heights[clade.clades[0]] + heights[clade.clades[-1]]
            ) / 2.0

        if tree.root.clades:
            calc_row(tree.root)
        return heights

    x_posns = get_x_positions(tree)
    y_posns = get_y_positions(tree)
    # The function draw_clade closes over the axes object
    axes = plt.gca()

    def draw_clade_lines(
        here, start, stop,
        use_linecollection=False,
        orientation="horizontal",
        color="black",
        lw=.1,
    ):
        """Create a line with or without a line collection object.

        Graphical formatting of the lines representing clades in the plot can be
        customized by altering this function.
        """
        if not use_linecollection and orientation == "horizontal":
            axes.hlines(here, start, stop, color=color, lw=lw)
        elif use_linecollection and orientation == "horizontal":
            horizontal_linecollections.append(
                mpcollections.LineCollection(
                    [[(start, here), (stop, here)]], color=color, lw=lw
                )
            )
        elif not use_linecollection and orientation == "vertical":
            axes.vlines(here, start, stop, color=color)
        elif use_linecollection and orientation == "vertical":
            vertical_linecollections.append(
                mpcollections.LineCollection(
                    [[(here, start), (here, stop)]], color=color, lw=lw
                )
            )

    def draw_clade(clade, x_start, color, lw):
        """Recursively draw a tree, down from the given clade."""
        x_here = x_posns[clade]
        y_here = y_posns[clade]
        # phyloXML-only graphics annotations
        if hasattr(clade, "color") and clade.color is not None:
            color = clade.color.to_hex()
        if hasattr(clade, "width") and clade.width is not None:
            lw = clade.width * lw
        # Draw a horizontal line from start to here
        orit = orientation
        draw_clade_lines(
            y_here, x_start, x_here,
            use_linecollection=True,
            orientation=orit,
            color=color,
            lw=lw,
        )
        # Add node/taxon labels
        if label_func is not None:
            label = label_func(clade)
            if label not in (None, clade.__class__.__name__):
                if orientation == 'horizontal':
                    x = x_here
                    y = y_here
                    va = 'center'
                    ha = 'right' if inverted else 'left'
                else:
                    x = y_here
                    y = x_here
                    ha = 'center'
                    va = 'top' if inverted else 'bottom'

                axes.text(x, y,
                    " %s" % label,
                    verticalalignment=va,
                    horizontalalignment=ha,
                    fontsize=fontsize,
                    color=get_label_color(label))

        # Add label above the branch (optional)
        conf_label = format_branch_label(clade)
        if conf_label:
            if orientation == 'horizontal':
                x = 0.5 * (x_start + x_here)
                y = y_here
                va = 'center'
                ha = 'right' if inverted else 'left'
            else:
                x = y_here
                y = 0.5 * (x_start + x_here)
                ha = 'center'
                va = 'top' if inverted else 'bottom'

            axes.text(x, y,
                conf_label,
                fontsize="small",
                verticalalignment=va,
                horizontalalignment=ha)

        if clade.clades:
            orit = 'horizontal' if orientation=='vertical' else 'vertical'
            # Draw a vertical line connecting all children
            y_top = y_posns[clade.clades[0]]
            y_bot = y_posns[clade.clades[-1]]
            # Only apply widths to horizontal lines, like Archaeopteryx
            draw_clade_lines(
                x_here, y_bot, y_top,
                use_linecollection=True,
                orientation=orit,
                color=color,
                lw=lw,
            )
            # Draw descendents
            for child in clade:
                draw_clade(child, x_here, color, lw)

    start = max(x_posns.values()) if inverted else 0.
    draw_clade(tree.root, start, "k", linewidth)

    # If line collections were used to create clade lines, here they are added
    # to the pyplot plot.
    for i in horizontal_linecollections:
        axes.add_collection(i)
    for i in vertical_linecollections:
        axes.add_collection(i)

    # Aesthetics

    try:
        name = tree.name
    except AttributeError:
        pass
    else:
        if name:
            axes.set_title(name)
    
    if orientation == 'horizontal':
        axes.set_xlabel("branch length")
        axes.set_ylabel("taxa")
    else:
        axes.set_ylabel("branch length")
        axes.set_xlabel("taxa")
        
    # Add margins around the tree to prevent overlapping the axes
    xmax = max(x_posns.values())

    if orientation == 'horizontal':
        if inverted:
            axes.set_xlim(-0.2 * xmax, 1.05 * xmax)
        else:
            axes.set_xlim(-0.05 * xmax, 1.2 * xmax)
        axes.set_ylim(0.2, max(y_posns.values()) + 0.8)
    else:
        if inverted:
            axes.set_ylim(-0.2 * xmax, 1.05 * xmax)
        else:
            axes.set_ylim(-0.05 * xmax, 1.2 * xmax)
        axes.set_xlim(0.2, max(y_posns.values()) + 0.8)

    # Parse and process key word arguments as pyplot options
    for key, value in kwargs.items():
        try:
            # Check that the pyplot option input is iterable, as required
            list(value)
        except TypeError:
            raise ValueError(
                'Keyword argument "%s=%s" is not in the format '
                "pyplot_option_name=(tuple), pyplot_option_name=(tuple, dict),"
                " or pyplot_option_name=(dict) " % (key, value)
            )
        if isinstance(value, dict):
            getattr(plt, str(key))(**dict(value))
        elif not (isinstance(value[0], tuple)):
            getattr(plt, str(key))(*value)
        elif isinstance(value[0], tuple):
            getattr(plt, str(key))(*value[0], **dict(value[1]))

    return x_posns, y_posns

class IndexFormatter(ticker.Formatter):
    """Function taken from Matplotlib version 3.3.1 as soon to be deprecated.

    Format the position x to the nearest i-th label where ``i = int(x + 0.5)``.
    Positions where ``i < 0`` or ``i > len(list)`` have no tick labels.

    :arg labels: List of labels.
    type labels: list
    """
    def __init__(self, labels):
        self.labels = labels
        self.n = len(labels)

    def __call__(self, x, pos=None):
        """Return the format for tick value *x* at position pos.

        The position is ignored and the value is rounded to the nearest
        integer, which is used to look up the label.
        """
        i = int(x + 0.5)
        if i < 0 or i >= self.n:
            return ''
        else:
            return self.labels[i]



from os import remove
import shlex
from os.path import isfile, join, split, splitext
from prody.tests import TestCase, skipIf, skipUnless

from numpy.testing import *

from prody.utilities import importDec
dec = importDec()

from prody import parsePDB, DCDFile, parseDCD

from prody.tests.datafiles import TEMPDIR, pathDatafile

from prody.apps import prody_parser

from prody.tests import MATPLOTLIB, NOPRODYCMD, WINDOWS

class TestCatdcdCommand(TestCase):

    def setUp(self):

        self.output = join(TEMPDIR, 'test_prody_catdcd.dcd')

        self.dcdpath = pathDatafile('dcd')
        self.pdbpath = pathDatafile('multi_model_truncated')

        self.dcd = DCDFile(self.dcdpath)
        self.ag = parsePDB(self.pdbpath, model=1)

        self.command = 'catdcd -o ' + self.output

        self.tearDown()

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testSimpleConcat(self):

        command = self.command + ' {0:s} {0:s} {0:s}'.format(self.dcdpath)

        namespace = prody_parser.parse_args(shlex.split(command))
        namespace.func(namespace)

        coords = self.dcd[:]._getCoordsets()
        concat = parseDCD(self.output)._getCoordsets()
        assert_equal(coords, concat[:3])
        assert_equal(coords, concat[3:6])
        assert_equal(coords, concat[6:])

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testSelectConcat(self):

        command = self.command + ' -s ca --pdb {1:s} {0:s} {0:s}'.format(
                                self.dcdpath, self.pdbpath)

        namespace = prody_parser.parse_args(shlex.split(command))
        namespace.func(namespace)

        select = self.ag.ca
        assert_equal(select.numAtoms(), 10)

        coords = self.dcd[:]
        coords.setAtoms(select)
        coords = coords._getCoordsets()

        concat = parseDCD(self.output)
        assert_equal(concat.numAtoms(), select.numAtoms())
        concat = concat._getCoordsets()

        assert_equal(select.numAtoms(), coords.shape[1])
        assert_equal(select.numAtoms(), concat.shape[1])
        assert_equal(coords, concat[:3])
        assert_equal(coords, concat[3:])

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testAlignConcat(self):

        command = self.command + ' --align ca --pdb {1:s} {0:s} {0:s}'.format(
                                self.dcdpath, self.pdbpath)

        namespace = prody_parser.parse_args(shlex.split(command))
        namespace.func(namespace)

        select = self.ag.ca

        coords = self.dcd[:]
        concat = parseDCD(self.output)

        assert_equal(concat.numAtoms(), coords.numAtoms())

        coords.setCoords(self.ag.getCoords())
        coords.setAtoms(select)
        coords.superpose()
        coords.setAtoms(None)
        coords = coords._getCoordsets()

        concat = concat._getCoordsets()

        assert_equal(coords, concat[:3])
        assert_equal(coords, concat[3:])

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command is not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testSelectException(self):

        command = self.command + ' -s ca {0:s} {0:s}'.format(
                                self.dcdpath)
        namespace = prody_parser.parse_args(shlex.split(command))
        self.assertRaises(ValueError, namespace.func, namespace)

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command is not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testAlignException(self):

        command = self.command + ' --align ca {0:s} {0:s}'.format(
                                self.dcdpath)
        namespace = prody_parser.parse_args(shlex.split(command))
        self.assertRaises(ValueError, namespace.func, namespace)

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command is not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testIOException(self):

        command = self.command + ' {0:s} {0:s}'.format('deneme.dcd')
        namespace = prody_parser.parse_args(shlex.split(command))
        self.assertRaises(IOError, namespace.func, namespace)

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command is not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testSelectException2(self):

        command = self.command + ' -s None {0:s} {0:s}'.format(self.dcdpath)
        namespace = prody_parser.parse_args(shlex.split(command))
        self.assertRaises(ValueError, namespace.func, namespace)


    def tearDown(self):

        if isfile(self.output): remove(self.output)



from prody.tests import TestCase

from prody.utilities import rangeString


class TestRangeString(TestCase):

    def testContinuous(self):

        self.assertEqual(rangeString(list(range(10))), '0 to 9')

    def testNegative(self):

        self.assertEqual(rangeString(list(range(-5, 10)), pos=False),
                         '-5 to 9')

    def testGapped(self):

        self.assertEqual(rangeString(list(range(-5, 10)) +
                                     list(range(15, 20)) +
                                     list(range(25, 30)), pos=False),
                                     '-5 to 9 15 to 19 25 to 29')

    def testRepeated(self):

        self.assertEqual(rangeString(list(range(10, 20)) +
                                     list(range(15, 20)) +
                                     list(range(30))), '0 to 29')



# -*- coding: utf-8 -*-
"""This module defines a class for handling multiple trajectories."""

import os.path

import numpy as np
from numbers import Integral

from .trajbase import TrajBase
from .frame import Frame

from prody.trajectory import openTrajFile

__all__ = ['Trajectory']

class Trajectory(TrajBase):

    """A class for handling trajectories in multiple files."""

    def __init__(self, name, **kwargs):
        """Trajectory can be instantiated with a *name* or a filename. When
        name is a valid path to a trajectory file it will be opened for
        reading."""

        TrajBase.__init__(self, name)
        self._trajectory = None
        self._trajectories = []
        self._filenames = set()
        self._n_files = 0
        self._cfi = 0 # current file index
        assert 'mode' not in kwargs, 'mode is an invalid keyword argument'
        self._kwargs = kwargs
        if os.path.isfile(name):
            self.addFile(name)

    def __repr__(self):

        if self._closed:
            return '<Trajectory: {0} (closed)>'.format(self._title)

        link = ''
        if self._ag is not None:
            link = 'linked to ' + str(self._ag) + '; '

        files = ''
        if self._n_files > 1:
            files = '{0} files; '.format(self._n_files)

        next = 'next {0} of {1} frames; '.format(self._nfi, self._n_csets)

        if self._indices is None:
            atoms = '{0} atoms'.format(self._n_atoms)
        else:
            atoms = 'selected {0} of {1} atoms'.format(
                                            self.numSelected(), self._n_atoms)

        return '<Trajectory: {0} ({1}{2}{3}{4})>'.format(
                                        self._title, link, files, next, atoms)

    def _nextFile(self):

        self._cfi += 1
        if self._cfi < self._n_files:
            self._trajectory = self._trajectories[self._cfi]
            if self._trajectory.nextIndex() > 0:
                self._trajectory.reset()

    def _gotoFile(self, i):

        if i < self._n_files:
            self._cfi = i
            self._trajectory = self._trajectories[i]
            if self._trajectory.nextIndex() > 0:
                self._trajectory.reset()

    def setAtoms(self, atoms):

        for traj in self._trajectories:
            traj.setAtoms(atoms)
        TrajBase.setAtoms(self, atoms)

    setAtoms.__doc__ = TrajBase.setAtoms.__doc__

    def link(self, *ag):

        if ag:
            TrajBase.link(self, *ag)
            for traj in self._trajectories:
                traj.link(*ag)
        else:
            return self._ag

    link.__doc__ = TrajBase.link.__doc__

    def addFile(self, filename, **kwargs):
        """Add a file to the trajectory instance. Currently only DCD files
        are supported."""

        if not isinstance(filename, str):
            raise ValueError('filename must be a string')
        if os.path.abspath(filename) in self._filenames:
            raise IOError('{0} is already added to the trajectory'
                          .format(filename))
        assert 'mode' not in kwargs, 'mode is an invalid keyword argument'
        traj = openTrajFile(filename, **(kwargs or self._kwargs))
        n_atoms = self._n_atoms
        if n_atoms != 0 and n_atoms != traj.numAtoms():
            raise IOError('{0} must have same number of atoms as '
                            'previously loaded files'.format(traj.getTitle()))

        if self._n_files == 0:
            self._trajectory = traj
            self._title = traj.getTitle()
        if n_atoms == 0:
            self._n_atoms = traj.numAtoms()
            self._coords = traj._coords
        self._trajectories.append(traj)
        self._n_csets += traj.numFrames()
        self._n_files += 1
        if self._ag is not None:
            traj.setAtoms(self._ag)

    def numFiles(self):
        """Returns number of open trajectory files."""

        return self._n_files

    def getFilenames(self, absolute=False):
        """Returns list of filenames opened for reading."""

        return [traj.getFilename(absolute) for traj in self._trajectories]

    def getFrame(self, index):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        self.goto(index)
        return next(self)

    def getCoordsets(self, indices=None):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if indices is None:
            indices = np.arange(self._n_csets)
        elif isinstance(indices, int):
            indices = np.array([indices])
        elif isinstance(indices, slice):
            indices = np.arange(*indices.indices(self._n_csets))
            indices.sort()
        elif isinstance(indices, (list, np.ndarray)):
            indices = np.unique(indices)
        else:
            raise TypeError('indices must be an integer or a list of '
                            'integers')

        nfi = self._nfi
        self.reset()
        coords = np.zeros((len(indices), self.numSelected(), 3),
                          self._trajectories[0]._dtype)
        prev = -1
        next = self.nextCoordset
        for i, index in enumerate(indices):
            diff = index - prev
            if diff > 1:
                self.skip(diff - 1)
            coords[i] = next()
            prev = index
        self.goto(nfi)
        return coords

    getCoordsets.__doc__ = TrajBase.getCoordsets.__doc__

    def __next__(self):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        nfi = self._nfi
        if nfi < self._n_csets:
            traj = self._trajectory
            while traj._nfi == traj._n_csets:
                self._nextFile()
                traj = self._trajectory
            unitcell = traj._nextUnitcell()
            coords = traj._nextCoordset()

            if self._ag is None:
                frame = Frame(self, nfi, coords, unitcell)
            else:
                frame = Frame(self, nfi, None, unitcell)
                self._ag.setACSLabel(self._title + ' frame ' + str(self._nfi))
            self._nfi += 1
            return frame

    __next__.__doc__ = TrajBase.__next__.__doc__
    next = __next__

    def nextCoordset(self):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if self._nfi < self._n_csets:
            traj = self._trajectory
            while traj._nfi == traj._n_csets:
                self._nextFile()
                traj = self._trajectory
            if self._ag is not None:
                self._ag.setACSLabel(self._title + ' frame ' + str(self._nfi))
                traj = self._trajectory
            self._nfi += 1
            #if self._indices is None:
            return traj.nextCoordset()
            #else:
            #    return traj.nextCoordset()[self._indices]

    nextCoordset.__doc__ = TrajBase.nextCoordset.__doc__

    def goto(self, n):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if not isinstance(n, Integral):
            raise ValueError('n must be an integer')
        n_csets = self._n_csets
        if n == 0:
            self.reset()
        else:
            if n < 0:
                n = n_csets + n
            if n < 0:
                n = 0
            elif n > n_csets:
                n = n_csets
            nfi = n
            for which, traj in enumerate(self._trajectories):
                if traj._n_csets >= nfi:
                    break
                else:
                    nfi -= traj._n_csets
            self._gotoFile(which)
            self._trajectory.goto(nfi)
            self._nfi = n

    goto.__doc__ = TrajBase.goto.__doc__

    def skip(self, n):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if not isinstance(n, Integral):
            raise ValueError('n must be an integer')
        left = self._n_csets - self._nfi
        if n > left:
            n = left
        while self._nfi < self._n_csets and n > 0:
            traj = self._trajectory
            skip = min(n, traj.numFrames() - traj.nextIndex())
            traj.skip(skip)
            if n > skip:
                self._nextFile()
            self._nfi += skip
            n -= skip

    skip.__doc__ = TrajBase.skip.__doc__

    def reset(self):

        if self._closed:
            raise ValueError('I/O operation on closed file')
        if self._trajectories:
            for traj in self._trajectories:
                traj.reset()
            self._trajectory = self._trajectories[0]
            self._cfi = 0
            self._nfi = 0

    reset.__doc__ = TrajBase.reset.__doc__

    def close(self):

        for traj in self._trajectories:
            traj.close()
        self._closed = True

    close.__doc__ = TrajBase.close.__doc__

    def hasUnitcell(self):

        return np.all([traj.hasUnitcell() for traj in self._trajectories])

    hasUnitcell.__doc__ = TrajBase.hasUnitcell.__doc__

    def getTimestep(self):
        """Returns list of timestep sizes, one number from each file."""

        return [traj.getTimestep() for traj in self._trajectories]

    def getFirstTimestep(self):
        """Returns list of first timestep values, one number from each file."""

        return [traj.getFirstTimestep() for traj in self._trajectories]

    def getFrameFreq(self):
        """Returns list of timesteps between frames, one number from each file.
        """

        return [traj.getFrameFreq() for traj in self._trajectories]

    def numFixed(self):
        """Returns a list of fixed atom numbers, one from each file."""

        return [traj.numFixed() for traj in self._trajectories]



import numpy as np
from scipy import sparse
from prody import LOGGER, SETTINGS
from prody import buildDistMatrix, calcDistFlucts

__all__ = ['calcSpectrusSims', 'MBSPointMutation', 'calcMBS', 'calcMBSfromSim']


def calcSpectrusSims(distFlucts, pdb, cutoff=10., sigma='MRSDF', **kwargs):
    
    coords = pdb.getCoords()
    n = coords.shape[0]
    
    if distFlucts.shape != (n, n):
        raise ValueError('distFlucts and atoms must have same linear '
                         'size (now %d and %d)' %(distFlucts.shape[0], n))

    # identify atom pairs within cutoff and store relative dist. flucts
    nearestNeighs = np.full((n, n), True, dtype=bool)
    np.fill_diagonal(nearestNeighs, False)
    if isinstance(cutoff, (int, float)):
        # compute inter-atomic distances
        dist = buildDistMatrix(coords)
        nearestNeighs &= (dist <= cutoff)
    elif cutoff is not None:
        raise ValueError('cutoff must be either a number or None. '
                         'Got: {0}'.format(type(cutoff)))
    nnDistFlucts = distFlucts[nearestNeighs]

    # set the sigma parameter for the Gaussian weights
    if sigma == 'MRSDF':
        # sigma is computed as the average of the root distance fluctuations
        # between residues within the distance cutoff, as defined in the 
        # SPECTRUS algorithm
        sigma = np.mean(np.sqrt(nnDistFlucts))
    elif sigma == 'RMSDF':
        # sigma is computed as the root mean squared dist. fluctuations
        # (faster to compute than MRSDF)
        sigma = np.sqrt(np.mean(nnDistFlucts))

    # check if sigma is a number
    try:
        ss = 2.*sigma**2
    except:
        raise ValueError('sigma must be \'MRSDF\', \'RMSDF\' or a number.')

    # compute the Gaussian weights only for residue pairs
    # within the distance cutoff
    reducedSims = np.where(nearestNeighs, np.exp(-distFlucts/ss), 0)
    np.fill_diagonal(reducedSims, 1.)
    sparseSims  = sparse.csr_matrix(reducedSims)
    sparse.csr_matrix.eliminate_zeros(sparseSims)
    
    return sparseSims, sigma


def MBSPointMutation(simMatrix, index, **kwargs):

    if sparse.issparse(simMatrix):
        # slightly faster
        newSim = simMatrix.tolil()
    else:
        newSim = simMatrix.copy()
    n = simMatrix.shape[0]

    # cut non-adjacent links around atom 'index'
    nonNearestNeighs = list(range(0,index-1)) + list(range(index+2,n))
    for j in nonNearestNeighs:
        newSim[index, j] = 0
        newSim[j, index] = 0
    return newSim


def _removeOutliers(data, Delta=100., **kwargs):
    assert Delta > 0.
    med = np.median(data[~np.isnan(data)])
    if not 0 <= med <= 1e+4:
        raise RuntimeError('MBS profile is not well defined: Check for ' +\
        'possible disconnected components in the PDB structure and remove them.')
    # compute median of |distances from median|
    dists = np.abs(data - med)
    mdev = np.median(dists[~np.isnan(dists)])
    # replace entries with 'nan' if outside the "safe"
    # interval (median - Delta*mdev, median + Delta*mdev)
    Delta_mdev = Delta * mdev
    for i, dist in enumerate(dists):
        if np.isnan(dist) or dist>Delta_mdev:
            data[i] = np.nan
    return data 


def calcMBSfromSim(simMatrix, nEvals=20, remove_outliers=True,
                   remove_offset=True, **kwargs):

    LOGGER.timeit('_MBS')
    n = simMatrix.shape[0]
    mbs = np.zeros(n) 
    for i in range(n):
        try:
            # cut "non-covalent" bonds around atom 'i'
            modSim = MBSPointMutation(simMatrix, i)
            # compute laplacian's spectrum of eigvals
            laplacian = sparse.csgraph.laplacian(modSim, normed=True)
            evals = sparse.linalg.eigsh(laplacian, k=min(nEvals, n-1), 
                                        which='SM', return_eigenvectors=False)
            # sort eigvals in ascending order
            evals = np.sort(evals)
            # compute MBS at site i
            mbs[i] = np.sum(1./evals[1:])
        except Exception as err:
            LOGGER.warn('Unable to compute MBS at position '
                        '{0}. {1}'.format(i, err))
            mbs[i] = np.nan
    if any(~np.isnan(mbs)):
        # remove outliers
        if remove_outliers is True:
            mbs = _removeOutliers(mbs, **kwargs)
        # remove offset
        if remove_offset is True:
            offset = min(mbs[~np.isnan(mbs)])
            mbs = mbs - offset 
    LOGGER.report('MBS computed in %.1fs.', '_MBS')

    return mbs


def calcMBS(anm, atomGroup, remove_outliers=True, remove_offset=True, 
            **kwargs):

    distFlucts = calcDistFlucts(anm, norm=False)
    sparseSim, sigma = calcSpectrusSims(distFlucts, atomGroup, **kwargs)
    mbs = calcMBSfromSim(sparseSim, **kwargs)

    return mbs






"""This module contains unit tests for :mod:`.atommap` module."""

from prody import *
from prody import LOGGER, SELECT
from prody.atomic import atommap
from prody.tests import unittest
from prody.tests.datafiles import *

prody.atomic.select.DEBUG = False
LOGGER.verbosity = 'none'

AG = parseDatafile('1ubi')

AM = AtomMap(AG, list(range(9, -1, -1)) + list(range(11, 20)),
             mapping=list(range(10)) + list(range(11, 20)), dummies=[10])


class TestInstantiation(unittest.TestCase):


    def testInstantiation(self):

        am = AtomMap(AG, list(range(9, -1, -1)) + [atommap.DUMMY] +
                     list(range(11, 20)),
                     dummies=True)

        self.assertEqual(AM, am)


class TestSelection(unittest.TestCase):

    def testAllSelection(self):

        self.assertEqual(AM, AM.all)

    def testChainSelection(self):

        self.assertEqual(AM, AM.select('chain A _'))

    def testNameSelection(self):

        self.assertEqual((AM.getNames() == 'CA').sum(),
                          len(AM.select('name CA')))

    def testGetBoolArray(self):

        self.assertEqual(len(AM), len(SELECT.getBoolArray(AM, 'name CA')))



"""This module defines functions for using compounds from the PDB and elsewhere."""

from .pdbligands import PDBLigandRecord
import numpy as np

__all__ = ['calc2DSimilarity', 'calc2DSimilarityMatrix']


def calc2DSimilarity(smiles1, smiles2):
    """Calculate 2D similarity using Morgan Fingerprints
    
    :arg smiles1: first SMILES string or PDBLigandRecord containing one
    :type smiles1: str, :class:`.PDBLigandRecord`

    :arg smiles2: second SMILES string or PDBLigandRecord containing one
    :type smiles2: str, :class:`.PDBLigandRecord`
    """
    try:
        from rdkit import Chem
        from rdkit import DataStructs
        from rdkit.Chem.Fingerprints import FingerprintMols
        from rdkit.Chem import AllChem
    except ImportError:
        raise ImportError('rdkit is a required package for calc2DSimilarity')

    if not isinstance(smiles1, str):
        if hasattr(smiles1, 'getCanonicalSMILES'):
            smiles1 = smiles1.getCanonicalSMILES()
        else:
            raise TypeError('smiles1 should be a string or an object with method getCanonicalSMILES')

    if not isinstance(smiles2, str):
        if hasattr(smiles2, 'getCanonicalSMILES'):
            smiles2 = smiles2.getCanonicalSMILES()
        else:
            raise TypeError('smiles2 should be a string or an object with method getCanonicalSMILES')

    m1 = Chem.MolFromSmiles(smiles1)
    m2 = Chem.MolFromSmiles(smiles2)
    if m1 is not None and m2 is not None:
        fp1 = AllChem.GetMorganFingerprint(m1, 2, useFeatures=True)
        fp2 = AllChem.GetMorganFingerprint(m2, 2, useFeatures=True)
        simi_score = DataStructs.TanimotoSimilarity(fp1, fp2)

    return simi_score


def calc2DSimilarityMatrix(smiles_set):
    num_smiles = len(smiles_set)
    sim_mat = np.ones((num_smiles, num_smiles))

    for i in range(num_smiles):
        for j in range(i+1, num_smiles):
            sim_mat[i, j] = sim_mat[j, i] = calc2DSimilarity(smiles_set[i], smiles_set[j])

    return sim_mat
    


"""Download PDB files for given identifiers."""

import os.path

from ..apptools import *

__all__ = ['prody_fetch']

def prody_fetch(*pdb, **kwargs):
    """Fetch PDB files from PDB FTP server.

    :arg pdbs: PDB identifier(s) or filename(s)

    :arg dir: target directory for saving PDB file(s), default is ``'.'``

    :arg gzip: gzip fetched files or not, default is **True**"""

    import prody


    pdblist = pdb
    if len(pdblist) == 1 and os.path.isfile(pdblist[0]):
        from prody.utilities import openFile
        with openFile(pdblist[0]) as inp:
            for item in inp.read().strip().split():
                for pdb in item.split(','):
                    if len(pdb) == 4 and pdb.isalnum():
                        pdblist.append(pdb)

    prody.fetchPDB(*pdblist, folder=kwargs.get('folder', '.'),
                   compressed=kwargs.get('gzip', False),
                   copy=True)

def addCommand(commands):

    subparser = commands.add_parser('fetch',
        help='fetch a PDB file')

    subparser.add_argument('--quiet', help="suppress info messages to stderr",
        action=Quiet, nargs=0)

    subparser.add_argument('--examples', action=UsageExample, nargs=0,
        help='show usage examples and exit')

    subparser.set_defaults(usage_example=
    """Download PDB file(s) by specifying identifiers:

  $ prody fetch 1mkp 1p38""",
    test_examples=[0]
    )

    subparser.add_argument('-d', '--dir', dest='folder', type=str,
        default='.', metavar='PATH',
        help=('target directory for saving PDB file(s)'))

    subparser.add_argument('-z', '--gzip', dest='gzip', action='store_true',
        default=False, help='write compressed PDB file(S)')

    subparser.add_argument('pdb', nargs='+',
        help='PDB identifier(s) or a file that contains them')

    subparser.set_defaults(func=lambda ns: prody_fetch(*ns.pdb, **ns.__dict__))
    subparser.set_defaults(subparser=subparser)



# -*- coding: utf-8 -*-
"""This module defines functions for accessing wwPDB servers."""

from os import getcwd
from glob import glob
from os.path import sep as pathsep
from os.path import isdir, isfile, join, split, splitext, normpath

from prody import LOGGER, SETTINGS
from prody.utilities import makePath, gunzip, relpath, copyFile, openURL
from prody.utilities import sympath, checkIdentifiers

__all__ = ['wwPDBServer', 'fetchPDBviaFTP', 'fetchPDBviaHTTP', 'WWPDB_FTP_SERVERS']


_WWPDB_RCSB = ('RCSB PDB (USA)', 'ftp.wwpdb.org', '/pub/')
_WWPDB_PDBe = ('PDBe (Europe)', 'ftp.ebi.ac.uk', '/pub/databases/rcsb/')
_WWPDB_PDBj = ('PDBj (Japan)', 'pdb.protein.osaka-u.ac.jp', '/pub/')

WWPDB_FTP_SERVERS = {
    'rcsb'   : _WWPDB_RCSB,
    'usa'    : _WWPDB_RCSB,
    'us'     : _WWPDB_RCSB,
    'pdbe'   : _WWPDB_PDBe,
    'euro'   : _WWPDB_PDBe,
    'europe' : _WWPDB_PDBe,
    'eu'     : _WWPDB_PDBe,
    'pdbj'   : _WWPDB_PDBj,
    'japan'  : _WWPDB_PDBj,
    'jp'     : _WWPDB_PDBj,
}

# _URL_US = lambda pdb: ('https://files.rcsb.org/pub/pdb/data/structures/all/pdb/pdb%s.ent.gz' %
                       # pdb.lower())
_URL_US = lambda pdb: ('https://files.rcsb.org/download/%s.pdb.gz' %
                       pdb.upper())
_URL_EU = lambda pdb: ('http://www.ebi.ac.uk/pdbe-srv/view/files/%s.ent.gz' %
                       pdb.lower())
_URL_JP = lambda pdb: ('http://www.pdbj.org/pdb_all/pdb%s.ent.gz' %
                       pdb.lower())
WWPDB_HTTP_URL = {
    'rcsb'   : _URL_US,
    'usa'    : _URL_US,
    'us'     : _URL_US,
    'pdbe'   : _URL_EU,
    'euro'   : _URL_EU,
    'europe' : _URL_EU,
    'eu'     : _URL_EU,
    'pdbj'   : _URL_JP,
    'japan'  : _URL_JP,
    'jp'     : _URL_JP,
}

def wwPDBServer(*key):
    """Set/get `wwPDB`_ FTP/HTTP server location used for downloading PDB
    structures.  Use one of the following keywords for setting a server:

    +---------------------------+-----------------------------+
    | wwPDB FTP server          | *Key* (case insensitive)    |
    +===========================+=============================+
    | RCSB PDB (USA) (default)  | RCSB, USA, US               |
    +---------------------------+-----------------------------+
    | PDBe (Europe)             | PDBe, Europe, Euro, EU      |
    +---------------------------+-----------------------------+
    | PDBj (Japan)              | PDBj, Japan, Jp             |
    +---------------------------+-----------------------------+

    .. _wwPDB: http://www.wwpdb.org/"""

    if not key:
        return SETTINGS.get('wwpdb', None)
    elif len(key) == 1:
        try:
            key = key[0].lower()
        except AttributeError:
            raise TypeError('key must be a string')
        if key in WWPDB_FTP_SERVERS:
            SETTINGS['wwpdb'] = key
            SETTINGS.save()
            LOGGER.info('wwPDB server is set to {}.'
                        .format(WWPDB_FTP_SERVERS[key][0]))
        else:
            raise ValueError('{0} is not a valid wwPDB server identifier'
                             .format(repr(key)))
    else:
        raise TypeError('one wwPDB server identifier is expected, {0} given'
                        .format(len(key)))



def fetchPDBviaFTP(*pdb, **kwargs):
    """Retrieve PDB (default), PDBML, mmCIF, or EMD file(s) for specified *pdb*
    identifier(s) and return path(s).  Downloaded files will be stored in
    local PDB folder, if one is set using :meth:`.pathPDBFolder`, and copied
    into *folder*, if specified by the user.  If no destination folder is
    specified, files will be saved in the current working directory.  If
    *compressed* is **False**, decompressed files will be copied into
    *folder*.  *format* keyword argument can be used to retrieve
    `PDBML <http://pdbml.pdb.org/>`_, `mmCIF <http://mmcif.pdb.org/>`_
    and `PDBML <ftp://ftp.wwpdb.org/pub/emdb/doc/Map-format/current/EMDB_map_format.pdf>`_ 
    files: ``format='cif'`` will fetch an mmCIF file, ``format='emd'`` will fetch an EMD file,
    and ``format='xml'`` will fetch a PDBML file. 
    If PDBML header file is desired, ``noatom=True`` argument will do the job."""

    format = str(kwargs.pop('format', 'pdb')).lower()
    if kwargs.get('check', True):
        identifiers = checkIdentifiers(*pdb, format=format)
    else:
        identifiers = list(pdb)

    output_folder = kwargs.pop('folder', None)
    compressed = bool(kwargs.pop('compressed', True))
    noatom = bool(kwargs.pop('noatom', False))

    if format == 'pdb':
        ftp_divided = 'pdb/data/structures/divided/pdb'
        ftp_pdbext = '.ent.gz'
        ftp_prefix = 'pdb'
        extension = '.pdb'
    elif format == 'xml':
        if noatom:
            ftp_divided = 'pdb/data/structures/divided/XML-noatom'
            ftp_pdbext = '-noatom.xml.gz'
            extension = '-noatom.xml'
        else:
            ftp_divided = 'pdb/data/structures/divided/XML'
            ftp_pdbext = '.xml.gz'
            extension = '.xml'
        ftp_prefix = ''
    elif format == 'cif':
        ftp_divided = 'pdb/data/structures/divided/mmCIF'
        ftp_pdbext = '.cif.gz'
        ftp_prefix = ''
        extension = '.cif'
    elif format == 'emd' or format == 'map':
        ftp_divided = 'emdb/structures'
        ftp_pdbext = '.map.gz'
        ftp_prefix = 'emd_'
        extension = '.map'
    else:
        raise ValueError(repr(format) + ' is not valid format')

    local_folder = pathPDBFolder()

    if format == 'pdb' and local_folder:
        local_folder, is_divided = local_folder
        if is_divided:
            getPath = lambda pdb: join(makePath(join(local_folder, pdb[1:3])),
                                       'pdb' + pdb + '.pdb.gz')
        else:
            getPath = lambda pdb: join(local_folder, pdb + '.pdb.gz')
        if output_folder is None:
            second = lambda filename, pdb: filename
        else:
            if compressed:
                second = lambda filename, pdb: (copyFile(filename,
                            join(output_folder, pdb + extension + '.gz')))
            else:
                second = lambda filename, pdb: gunzip(filename,
                            join(output_folder, pdb + extension))

    else:
        if output_folder is None:
            output_folder = getcwd()
        if compressed:
            getPath = lambda pdb: join(output_folder, pdb + extension + '.gz')
            second = lambda filename, pdb: filename
        else:
            getPath = lambda pdb: join(output_folder, pdb + extension)
            second = lambda filename, pdb: gunzip(getPath(pdb), getPath(pdb))


    ftp_name, ftp_host, ftp_path = WWPDB_FTP_SERVERS[wwPDBServer() or 'us']
    LOGGER.debug('Connecting wwPDB FTP server {0}.'.format(ftp_name))

    from ftplib import FTP
    try:
        ftp = FTP(ftp_host)
    except Exception as error:
        raise type(error)('FTP connection problem, potential reason: '
                          'no internet connectivity')
    else:
        success = 0
        failure = 0
        filenames = []
        ftp.login('')
        for pdb in identifiers:
            if pdb is None:
                filenames.append(None)
                continue
            data = []
            ftp_fn = ftp_prefix + pdb + ftp_pdbext
            try:
                ftp.cwd(ftp_path)
                ftp.cwd(ftp_divided)
                if format == 'emd':
                    ftp.cwd('EMD-{0}/map'.format(pdb))
                else:
                    ftp.cwd(pdb[1:3])
                ftp.retrbinary('RETR ' + ftp_fn, data.append)
            except Exception as error:
                if ftp_fn in ftp.nlst():
                    LOGGER.warn('{0} download failed ({1}). It is '
                                'possible that you do not have rights to '
                                'download .gz files in the current network.'
                                .format(pdb, str(error)))
                else:
                    LOGGER.info('{0} download failed. {1} does not exist '
                                'on {2}.'.format(ftp_fn, pdb, ftp_host))
                failure += 1
                filenames.append(None)
            else:
                if len(data):
                    filename = getPath(pdb)

                    with open(filename, 'w+b') as pdbfile:
                        write = pdbfile.write
                        [write(block) for block in data]

                    filename = normpath(relpath(second(filename, pdb)))
                    LOGGER.debug('{0} downloaded ({1})'
                                 .format(pdb, sympath(filename)))
                    success += 1
                    filenames.append(filename)
                else:
                    LOGGER.warn('{0} download failed, reason unknown.'
                                .format(pdb))
                    failure += 1
                    filenames.append(None)

        ftp.quit()

    LOGGER.debug('PDB download via FTP completed ({0} downloaded, '
                 '{1} failed).'.format(success, failure))
    if len(identifiers) == 1:
        return filenames[0]
    else:
        return filenames


def fetchPDBviaHTTP(*pdb, **kwargs):
    """Retrieve PDB file(s) for specified *pdb* identifier(s) and return
    path(s).  Downloaded files will be stored in local PDB folder, if one
    is set using :meth:`.pathPDBFolder`, and copied into *folder*, if
    specified by the user.  If no destination folder is specified, files
    will be saved in the current working directory.  If *compressed* is
    **False**, decompressed files will be copied into *folder*."""

    if kwargs.get('check', True):
        identifiers = checkIdentifiers(*pdb)
    else:
        identifiers = list(pdb)

    output_folder = kwargs.pop('folder', None)
    compressed = bool(kwargs.pop('compressed', True))

    format = kwargs.get('format', 'pdb')
    if (eval(long_id_check_str % 12) or eval(long_id_check_str % 13)):
        format = 'cif'

    noatom = bool(kwargs.pop('noatom', False))
    if format == 'pdb':
        extension = '.pdb'
    elif format == 'xml':
        if noatom:
            extension = '-noatom.xml'
        else:
            extension = '.xml'
    elif format == 'cif':
        extension = '.cif'
    elif format == 'emd' or format == 'map':
        extension = '.map'
    else:
        raise ValueError(repr(format) + ' is not valid format')

    local_folder = pathPDBFolder()
    if local_folder:
        local_folder, is_divided = local_folder
        if is_divided:
            getPath = lambda pdb: join(makePath(join(local_folder, pdb[1:3])),
                                       'pdb' + pdb + '.pdb.gz')
        else:
            getPath = lambda pdb: join(local_folder, pdb + '.pdb.gz')
        if output_folder is None:
            second = lambda filename, pdb: filename
        else:
            if compressed:
                second = lambda filename, pdb: (copyFile(filename,
                            join(output_folder, pdb + extension + '.gz')))
            else:
                second = lambda filename, pdb: gunzip(filename,
                            join(output_folder, pdb + extension))

    else:
        if output_folder is None:
            output_folder = getcwd()
        if compressed:
            getPath = lambda pdb: join(output_folder, pdb + extension + '.gz')
            second = lambda filename, pdb: filename
        else:
            getPath = lambda pdb: join(output_folder, pdb + extension)
            second = lambda filename, pdb: gunzip(getPath(pdb), getPath(pdb))


    getURL = WWPDB_HTTP_URL[wwPDBServer() or 'us']

    success = 0
    failure = 0
    filenames = []
    for pdb in identifiers:
        if pdb is None:
            filenames.append(None)
            continue
        try:
            url = getURL(pdb)
            if kwargs.get('format', 'pdb') != 'pdb':
                url = url.replace('.pdb', extension)
            handle = openURL(url)
        except Exception as err:
            if not (eval(long_id_check_str % 12) or eval(long_id_check_str % 13)):
                LOGGER.warn('{0} download failed ({1}).'.format(pdb, str(err)))
            failure += 1
            filenames.append(None)
        else:
            data = handle.read()
            if len(data):
                filename = getPath(pdb)

                with open(filename, 'w+b') as pdbfile:
                    pdbfile.write(data)

                filename = normpath(relpath(second(filename, pdb)))
                LOGGER.debug('{0} downloaded ({1})'
                             .format(pdb, sympath(filename)))
                success += 1
                filenames.append(filename)
            else:
                LOGGER.warn('{0} download failed, reason unknown.'
                            .format(pdb))
                failure += 1
                filenames.append(None)
    LOGGER.debug('PDB download via HTTP completed ({0} downloaded, '
                 '{1} failed).'.format(success, failure))
    if len(identifiers) == 1:
        return filenames[0]
    else:
        return filenames

if __name__ == '__main__':

    pdbids = ['1mkp', '1zz2', 'nano']

    for gzip in [False, True]:
        fetchPDBviaFTP(*pdbids, compressed=gzip, folder='.')
        fetchPDBviaFTP(*pdbids, compressed=gzip, folder='.', format='cif')
        fetchPDBviaFTP(*pdbids, compressed=gzip, folder='.', format='xml')
        fetchPDBviaFTP(*pdbids, compressed=gzip, folder='.', format='xml',
                       noatom=1)
        fetchPDBviaHTTP(*pdbids, compressed=gzip, folder='.')
    from glob import glob
    from os import remove
    for pdb in pdbids:
        fns = glob(pdb + '.*')
        print((pdb, '>', ', '.join(fns)))
        for fn in fns:
            remove(fn)



# -*- coding: utf-8 -*-
"""This module defines a class and a function for using the tensorial network model, 
which includes bond-bending and twist elasticities."""

import numpy as np

from prody import LOGGER
from prody.atomic import Atomic, AtomGroup
from prody.proteins import parsePDB
from prody.utilities import importLA, checkCoords

from .anm import ANMBase
from .gnm import GNMBase, ZERO, checkENMParameters
from numpy import eye, arccos, zeros, linalg, dot, tan, sqrt, pi


__all__ = ['bbENM']

class Increment(object):

    def __init__(self, s=0):

        self._i = s

    def __call__(self, i=1):

        self._i += i
        return self._i


class bbENM(ANMBase):

    """Class for bond bending correction for anisotropic network model ([AS12]_).
    Additional argument is the b parameter which is the bond bending constant in the model. 
    .. [AS12] Srivastava A, Halevi RB, Veksler A, Granek, R. Tensorial elastic network model 
    for protein dynamics: Integration of the anisotropic model with bond-bending and twist
    elasticities.
       *Proteins* **2012** 80:2692-2700.

    """

    def __init__(self, name='Unknown'):

        super(bbENM, self).__init__(name)


    def buildHessian(self, coords, B=1., cutoff=15., gamma=1., **kwargs):
        """Build Hessian matrix for given coordinate set.

        :arg coords: a coordinate set or an object with ``getCoords`` method
        :type coords: :class:`numpy.ndarray`

        :arg b: bond-bending constant, default is 1.0
        :type b: float

        :arg cutoff: cutoff distance (Å) for pairwise interactions,
            default is 15.0 Å
        :type cutoff: float

        :arg gamma: spring constant, default is 1.0
        :type gamma: float

        :type scale: float

        """

        try:
            coords = (coords._getCoords() if hasattr(coords, '_getCoords') else
                      coords.getCoords())
        except AttributeError:
            try:
                checkCoords(coords)
            except TypeError:
                raise TypeError('coords must be a Numpy array or an object '
                                'with `getCoords` method')

        LOGGER.timeit('_bbenm')
        self._n_atoms = natoms = int(coords.shape[0])

        self._hessian = hessian = np.zeros((3*natoms, 3*natoms), float)
        self._dof = 3*natoms - 6
        
        # anm hessian calculation 
        cutoff, gamma, gamma_func = checkENMParameters(cutoff, gamma)
        cutoff2 = cutoff * cutoff
        for i in range(natoms):
            res_i3 = i*3
            res_i33 = res_i3+3
            i_p1 = i+1
            i2j_all = coords[i_p1:, :] - coords[i]
            for j, dist2 in enumerate((i2j_all ** 2).sum(1)):
                if dist2 > cutoff2:
                    continue
                i2j = i2j_all[j]
                j += i_p1
                g = gamma_func(dist2, i, j)
                res_j3 = j*3
                res_j33 = res_j3+3
                super_element = np.outer(i2j, i2j) * (- g / dist2)
                hessian[res_i3:res_i33, res_j3:res_j33] = super_element
                hessian[res_j3:res_j33, res_i3:res_i33] = super_element
                hessian[res_i3:res_i33, res_i3:res_i33] = \
                    hessian[res_i3:res_i33, res_i3:res_i33] - super_element
                hessian[res_j3:res_j33, res_j3:res_j33] = \
                    hessian[res_j3:res_j33, res_j3:res_j33] - super_element

        # hessian updates
        from .bbenmtools import buildhessian

        buildhessian(coords, hessian, natoms, 
                     float(cutoff), float(gamma),)

        LOGGER.report('Hessian was built in %.2fs.', label='_bbenm')

    def calcModes(self, n_modes=20, zeros=False, turbo=True):
         """Calculate normal modes.  This method uses :func:`scipy.linalg.eigh`
         function to diagonalize the Hessian matrix. When Scipy is not found,
         :func:`numpy.linalg.eigh` is used.

         :arg n_modes: number of non-zero eigenvalues/vectors to calculate.
             If **None** is given, all modes will be calculated.
         :type n_modes: int or None, default is 20

         :arg zeros: If **True**, modes with zero eigenvalues will be kept.
         :type zeros: bool, default is **True**

         :arg turbo: Use a memory intensive, but faster way to calculate modes.
         :type turbo: bool, default is **True**
         """

         super(bbENM, self).calcModes(n_modes, zeros, turbo)

    #     self._array = np.dot(self._project, self._array)


def test(pdb='2ci2'):

    from prody import parsePDB
    from numpy import zeros

    pdb = parsePDB(pdb, subset='ca')
    bbenm = bbENM()
    bbenm.buildHessian(pdb, cutoff=12.)
    return bbenm




# """This module contains unit tests for :mod:`prody.database.pfam` module."""

from prody.tests import unittest
from prody.database.pfam import searchPfam
from prody.database.pfam import fetchPfamMSA
from prody.database.pfam import parsePfamPDBs

from prody.atomic.selection import Selection

import os
import shutil

from prody import LOGGER
LOGGER.verbosity = 'none'

class TestSearchPfam(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.workdir = 'pfam_search_tests'
        if not os.path.exists(cls.workdir):
            os.mkdir(cls.workdir)
        os.chdir(cls.workdir)

        cls.queries = ['P19491', '6qkcB', '6qkcI', 'PF00047',
                       'hellow', 'hello']

    def testUniprotAccMulti(self):
        """Test the outcome of a simple search scenario using a Uniprot Accession
        for a multi-domain protein, AMPAR GluA2."""

        a = searchPfam(self.queries[0])

        self.assertIsInstance(a, dict,
            'searchPfam failed to return a dict instance')
        
        self.assertEqual(sorted(list(a.keys())),
                         ['PF00060', 'PF01094', 'PF10613'],
                         'searchPfam failed to return the right domain family IDs')
        
    def testPdbIdChMulti(self):
        """Test the outcome of a simple search scenario using a PDB ID
        and chain ID for the same multi-domain protein from specifying chain B."""

        a = searchPfam(self.queries[1])

        self.assertIsInstance(a, dict,
            'searchPfam failed to return a dict instance')
        
        self.assertEqual(sorted(list(a.keys())), ['PF00060', 'PF01094', 'PF10613'],
                         'searchPfam failed to return the right domain family IDs for AMPAR')

    def testPdbIdChSingle(self):
        """Test the outcome of a simple search scenario using a PDB ID
        and chain ID to get the single domain protein TARP g8 from chain I."""

        a = searchPfam(self.queries[2])

        self.assertIsInstance(a, dict,
            'searchPfam failed to return a dict instance')

        self.assertEqual(sorted(list(a.keys())),
                         ['PF00822', 'PF13903'],
                         'searchPfam failed to return the right domain family IDs for TARP')

    def testPfamInput(self):
        """Test the outcome of a search scenario where a Pfam ID is
        provided as input."""

        a = searchPfam(self.queries[3])

        self.assertIsInstance(a, dict,
            'searchPfam failed to return None for Pfam ID input {0}'.format(self.queries[3]))

    def testWrongInput1(self):
        """Test the outcome of a search scenario where a 6-char text is
        provided as input."""

        with self.assertRaises(OSError):
            searchPfam(self.queries[4])

    def testWrongInput2(self):
        """Test the outcome of a search scenario where a 5-char text is
        provided as input."""

        with self.assertRaises(ValueError):
            searchPfam(self.queries[5])

    @classmethod
    def tearDownClass(cls):
        os.chdir('..')
        shutil.rmtree(cls.workdir)


class TestFetchPfamMSA(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.query = 'PF00822'

        cls.workdir = 'pfam_msa_tests'
        if not os.path.exists(cls.workdir):
            os.mkdir(cls.workdir)
        os.chdir(cls.workdir)

    def testDefault(self):
        """Test the outcome of fetching the domain MSA for claudins
        with default parameters."""

        b = fetchPfamMSA(self.query)

        self.assertIsInstance(b, str,
            'fetchPfamMSA failed to return a str instance')
        
        self.assertEqual(b, 'PF00822_seed.sth')
        
        self.assertTrue(os.path.exists(b))


    def testSeed(self):
        """Test the outcome of fetching the domain MSA for claudins
        with the alignment type argument set to seed"""

        b = fetchPfamMSA(self.query, "seed")

        self.assertIsInstance(b, str,
            'fetchPfamMSA failed to return a str instance')
        
        self.assertEqual(b, 'PF00822_seed.sth')
        
        self.assertTrue(os.path.exists(b))

    def testFolder(self):
        """Test the outcome of fetching the domain MSA for claudins
        with keyword folder set to a folder that is made especially."""

        folder = "new_folder"
        os.mkdir(folder)
        b = fetchPfamMSA(self.query, folder=folder)

        self.assertIsInstance(b, str,
            'fetchPfamMSA failed to return a str instance')
        
        self.assertEqual(b, 'new_folder/PF00822_seed.sth')
        
        self.assertTrue(os.path.exists(b))
    
    @classmethod
    def tearDownClass(cls):
        os.chdir('..')
        shutil.rmtree(cls.workdir)
            

class TestParsePfamPDBs(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.queries = ['PF20446', 'Q57ZF2', 'P40682']

        cls.workdir = 'pfam_pdb_tests'
        if not os.path.exists(cls.workdir):
            os.mkdir(cls.workdir)
        os.chdir(cls.workdir)

    def testPfamIdDefault(self):
        """Test the outcome of parsing PDBs for a tiny family
        of ABC class ATPase N-terminal domains (5 members)
        with the Pfam ID and default parameters."""

        b = parsePfamPDBs(self.queries[0])

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(len(b), 5,
            'parsePfamPDBs failed to return a list of length 5')


    def testUniprotDefault(self):
        """Test the outcome of parsing PDBs for a tiny family
        of ABC class ATPase N-terminal domains (5 members)
        with the Uniprot long ID and default parameters."""

        b = parsePfamPDBs(self.queries[1])

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(len(b), 5,
            'parsePfamPDBs failed to return a list of length 5')

        
    def testMultiDomainDefault(self):
        """Test the outcome of parsing PDBs using a V-type proton ATPase subunit S1,
        which has two domains but few relatives. Default parameters should
        return Selection objects containing the first domain."""

        b = parsePfamPDBs(self.queries[2])

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(b[0].getResnums()[0], 264,
            'parsePfamPDBs failed to return a first Selection with first resnum 264')

    def testMultiDomainStart1(self):
        """Test the outcome of parsing PDBs using a V-type proton ATPase subunit S1,
        which has two domains but few relatives. Using start=1 should be like default and
        return Selection objects containing the first domain."""

        b = parsePfamPDBs(self.queries[2], start=1)

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(b[0].getResnums()[0], 264,
            'parsePfamPDBs failed to return a first Selection with first resnum 264')
        
    def testMultiDomainStart2(self):
        """Test the outcome of parsing PDBs using a V-type proton ATPase subunit S1,
        which has two domains but few relatives. Setting start to 418 should
        return Selection objects containing the second domain."""

        b = parsePfamPDBs(self.queries[2], start=418)

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(b[0].getResnums()[0], 217,
            'parsePfamPDBs failed to return a first Selection with first resnum 217')

    def testPfamIdNumPdbs(self):
        """Test the outcome of parsing PDBs for a tiny family
        of ABC class ATPase N-terminal domains (5 members)
        with the Pfam ID and default parameters."""

        b = parsePfamPDBs(self.queries[0], num_pdbs=2)

        self.assertIsInstance(b, list,
            'parsePfamPDBs failed to return a list instance')

        self.assertIsInstance(b[0], Selection,
            'parsePfamPDBs failed to return a list of Selection instances')
        
        self.assertEqual(len(b), 2,
            'parsePfamPDBs failed to return a list of length 2 with num_pdbs=2')

    @classmethod
    def tearDownClass(cls):
        os.chdir('..')
        shutil.rmtree(cls.workdir)




# -*- coding: utf-8 -*-
"""This module defines functions for fetching PDB ligand data."""

from os.path import isdir, isfile, join, split, splitext

import numpy as np

from prody import LOGGER, SETTINGS, getPackagePath, PY3K
from prody.atomic import AtomGroup, ATOMIC_FIELDS
from prody.utilities import openFile, makePath, openURL
from .ccd import parseCCD

__all__ = ['PDBLigandRecord', 'fetchPDBLigand', 'parsePDBLigand']


class PDBLigandRecord(object):
    """Class for handling the output of fetchPDBLigand"""
    def __init__(self, data):
        self._rawdata = data

    def getCanonicalSMILES(self):
        canonical = None
        for row in self._rawdata[2].data:
            if row['_pdbx_chem_comp_descriptor.type'] == 'SMILES_CANONICAL':
                canonical = row['_pdbx_chem_comp_descriptor.descriptor']
        return canonical


def fetchPDBLigand(cci, filename=None):
    """Handle PDB ligand data from PDB_ for chemical component *cci*.
    *cci* may be 3-letter chemical component identifier or a valid XML
    filename. If *filename* is given, XML file will be saved with that name.

    This function may not work as _PDB is not hosting ligand XML files anymore.
    It is kept for use with existing ligand XML files.

    If you query ligand data frequently, you may configure ProDy to save XML
    files in your computer.  Set ``ligand_xml_save`` option **True**, i.e.
    ``confProDy(ligand_xml_save=True)``.  Compressed XML files will be save
    to ProDy package folder, e.g. :file:`/home/user/.prody/pdbligands`.  Each
    file is around 5Kb when compressed.

    This function is compatible with PDBx/PDBML v 4.0.

    Ligand data is returned in a dictionary.  Ligand coordinate atom data with
    *model* and *ideal* coordinate sets are also stored in this dictionary.
    Note that this dictionary will contain data that is present in the XML
    file and all Ligand Expo XML files do not contain every possible data
    field.  So, it may be better if you use :meth:`dict.get` instead of
    indexing the dictionary, e.g. to retrieve formula weight (or relative
    molar mass) of the chemical component use ``data.get('formula_weight')``
    instead of ``data['formula_weight']`` to avoid exceptions when this data
    field is not found in the XML file.  URL and/or path of the XML file are
    returned in the dictionary with keys ``url`` and ``path``, respectively.

    Following example downloads data for ligand STI (a.k.a. Gleevec and
    Imatinib) and calculates RMSD between model (X-ray structure 1IEP) and
    ideal (energy minimized) coordinate sets:

    .. ipython:: python
       :okexcept:

       from prody import *
       ligand_data = fetchPDBLigand('STI')
       ligand_data['model_coordinates_db_code']
       ligand_model = ligand_data['model']
       ligand_ideal = ligand_data['ideal']
       transformation = superpose(ligand_ideal.noh, ligand_model.noh)
       calcRMSD(ligand_ideal.noh, ligand_model.noh)"""

    if not isinstance(cci, str):
        raise TypeError('cci must be a string')

    if isfile(cci):
        inp = openFile(cci)
        xml = inp.read()
        inp.close()
        url = None
        path = cci
        cci = splitext(splitext(split(cci)[1])[0])[0].upper()
    elif len(cci) > 4 or not cci.isalnum():
        raise ValueError('cci must be 3-letters long and alphanumeric or '
                         'a valid filename')
    else:
        xml = None
        cci = cci.upper()
        if SETTINGS.get('ligand_xml_save'):
            folder = join(getPackagePath(), 'pdbligands')
            if not isdir(folder):
                makePath(folder)
            xmlgz = path = join(folder, cci + '.xml.gz')
            if isfile(xmlgz):
                with openFile(xmlgz) as inp:
                    xml = inp.read()
        else:
            folder = None
            path = None

        url = ('http://files.rcsb.org/ligands/download/{0}'
               '.xml'.format(cci.upper()))
        if not xml:
            try:
                inp = openURL(url)
            except IOError:
                raise IOError('XML file for ligand {0} is not found online'
                              .format(cci))
            else:
                xml = inp.read()
                if PY3K:
                    xml = xml.decode()
                inp.close()

            if filename:
                out = openFile(filename, mode='w', folder=folder)
                out.write(xml)
                out.close()
            if SETTINGS.get('ligand_xml_save'):
                with openFile(xmlgz, 'w') as out:
                    out.write(xml)

    import xml.etree.cElementTree as ET

    root = ET.XML(xml)
    if (root.get('{http://www.w3.org/2001/XMLSchema-instance}'
                 'schemaLocation') !=
            'http://pdbml.pdb.org/schema/pdbx-v40.xsd pdbx-v40.xsd'):
        LOGGER.warn('XML is not in PDBx/PDBML v 4.0 format, resulting '
                    'dictionary may not contain all data fields')
    ns = root.tag[:root.tag.rfind('}')+1]
    len_ns = len(ns)
    dict_ = {'url': url, 'path': path}

    for child in list(root.find(ns + 'chem_compCategory')[0]):
        tag = child.tag[len_ns:]
        if tag.startswith('pdbx_'):
            tag = tag[5:]
        dict_[tag] = child.text
    dict_['formula_weight'] = float(dict_.get('formula_weight'))

    identifiers_and_descriptors = []
    results = root.find(ns + 'pdbx_chem_comp_identifierCategory')
    if results:
        identifiers_and_descriptors.extend(results)
    results = root.find(ns + 'pdbx_chem_comp_descriptorCategory')
    if results:
        identifiers_and_descriptors.extend(results)
    for child in identifiers_and_descriptors:
        program = child.get('program').replace(' ', '_')
        type_ = child.get('type').replace(' ', '_')
        dict_[program + '_' + type_] = child[0].text
        dict_[program + '_version'] = child.get('program_version')

    dict_['audits'] = [(audit.get('action_type'), audit.get('date'))
                       for audit in
                       list(root.find(ns + 'pdbx_chem_comp_auditCategory'))]

    atoms = list(root.find(ns + 'chem_comp_atomCategory'))
    n_atoms = len(atoms)
    ideal_coords = np.zeros((n_atoms, 3))
    model_coords = np.zeros((n_atoms, 3))

    atomnames = np.zeros(n_atoms, dtype=ATOMIC_FIELDS['name'].dtype)
    elements = np.zeros(n_atoms, dtype=ATOMIC_FIELDS['element'].dtype)
    resnames = np.zeros(n_atoms, dtype=ATOMIC_FIELDS['resname'].dtype)
    charges = np.zeros(n_atoms, dtype=ATOMIC_FIELDS['charge'].dtype)

    resnums = np.ones(n_atoms, dtype=ATOMIC_FIELDS['charge'].dtype)

    alternate_atomnames = np.zeros(n_atoms, dtype=ATOMIC_FIELDS['name'].dtype)
    leaving_atom_flags = np.zeros(n_atoms, bool)
    aromatic_flags = np.zeros(n_atoms, bool)
    stereo_configs = np.zeros(n_atoms, bool)
    ordinals = np.zeros(n_atoms, int)

    name2index = {}

    for i, atom in enumerate(atoms):
        data = dict([(child.tag[len_ns:], child.text) for child in list(atom)])

        name = data.get('pdbx_component_atom_id', 'X')
        name2index[name] = i
        atomnames[i] = name
        elements[i] = data.get('type_symbol', 'X')
        resnames[i] = data.get('pdbx_component_comp_id', 'UNK')
        charges[i] = float(data.get('charge', 0))

        alternate_atomnames[i] = data.get('alt_atom_id', 'X')
        leaving_atom_flags[i] = data.get('pdbx_leaving_atom_flag') == 'Y'
        aromatic_flags[i] = data.get('pdbx_atomatic_flag') == 'Y'
        stereo_configs[i] = data.get('pdbx_stereo_config') == 'Y'
        ordinals[i] = int(data.get('pdbx_ordinal', 0))

        model_coords[i, 0] = float(data.get('model_Cartn_x', 0))
        model_coords[i, 1] = float(data.get('model_Cartn_y', 0))
        model_coords[i, 2] = float(data.get('model_Cartn_z', 0))
        ideal_coords[i, 0] = float(data.get('pdbx_model_Cartn_x_ideal', 0))
        ideal_coords[i, 1] = float(data.get('pdbx_model_Cartn_y_ideal', 0))
        ideal_coords[i, 2] = float(data.get('pdbx_model_Cartn_z_ideal', 0))

    pdbid = dict_.get('model_coordinates_db_code')
    if pdbid:
        model = AtomGroup(cci + ' model ({0})'.format(pdbid))
    else:
        model = AtomGroup(cci + ' model')
    model.setCoords(model_coords)
    model.setNames(atomnames)
    model.setResnames(resnames)
    model.setResnums(resnums)
    model.setElements(elements)
    model.setCharges(charges)
    model.setFlags('leaving_atom_flags', leaving_atom_flags)
    model.setFlags('aromatic_flags', aromatic_flags)
    model.setFlags('stereo_configs', stereo_configs)
    model.setData('ordinals', ordinals)
    model.setData('alternate_atomnames', alternate_atomnames)
    dict_['model'] = model
    ideal = model.copy()
    ideal.setTitle(cci + ' ideal')
    ideal.setCoords(ideal_coords)
    dict_['ideal'] = ideal

    bonds = []
    warned = set()
    for bond in list(root.find(ns + 'chem_comp_bondCategory') or bonds):
        name_1 = bond.get('atom_id_1')
        name_2 = bond.get('atom_id_2')
        try:
            bonds.append((name2index[name_1], name2index[name_2]))
        except KeyError:
            if name_1 not in warned and name_1 not in name2index:
                warned.add(name_1)
                LOGGER.warn('{0} specified {1} in bond category is not '
                            'a valid atom name.'.format(repr(name_1), cci))
            if name_2 not in warned and name_2 not in name2index:
                warned.add(name_2)
                LOGGER.warn('{0} specified {1} in bond category is not '
                            'a valid atom name.'.format(repr(name_2), cci))
    if bonds:
        bonds = np.array(bonds, int)
        model.setBonds(bonds)
        ideal.setBonds(bonds)
    return dict_


def parsePDBLigand(cci):
    """See :func:`.fetchPDBLigand`"""
    lig_dict = parseCCD(cci)
    return PDBLigandRecord(lig_dict)




# -*- coding: utf-8 -*-
"""This module defines functions for supporting VMD plugin `Heatmapper`_
format files.

.. _Heatmapper: http://www.ks.uiuc.edu/Research/vmd/plugins/heatmapper/"""

__author__ = 'Ahmet Bakan, Anindita Dutta'

from numpy import arange, fromstring, frombuffer, array

from prody.utilities import openFile, intorfloat, startswith, addext

from prody import LOGGER


__all__ = ['parseHeatmap', 'writeHeatmap', 'showHeatmap']

HMTYPES = {
    'max': float,
    'min': float,
    'numbering': lambda line: line.split(':'),
    'title': str,
    'xlabel': str,
    'ylabel': str,
    'xorigin': intorfloat,
    'xstep': intorfloat,
}


def showHeatmap(heatmap, *args, **kwargs):
    """Show *heatmap*, which can be an two dimensional array or a Heat Mapper
    :file:`.hm` file.

    Heatmap is plotted using :func:`~matplotlib.pyplot.imshow` function.
    Default values passed to this function are ``interpolation='nearest'``,
    ``aspect='auto'``, and ``origin='lower'``."""

    try:
        ndim, shape = heatmap.ndim, heatmap.shape
    except AttributeError:
        heatmap, headers = parseHeatmap(heatmap)
        ndim, shape = heatmap.ndim, heatmap.shape

        xorigin = headers.pop('xorigin', 0)
        xextent = headers.pop('xstep', 1) * shape[0]

        ylabel = kwargs.get('ylabel', '').lower()
        indices = None
        if ylabel:
            for key in headers.get('numbering', []):
                if startswith(ylabel, key.lower()):
                    indices = headers.get(key)
        if indices is not None:
            extent = [indices[0] - .5, indices[0] + len(indices) - .5,
                      xorigin - .5, xextent - .5]
        else:
            extent = [-.5, shape[1] * 2 - .5, xorigin - .5, xextent - .5]
        kwargs.setdefault('extent', extent)

        for key in ['numbering', 'min', 'max'] + headers.get('numbering', []):
            headers.pop(key, None)
        headers.update(kwargs)
        kwargs = headers


    if ndim != 2:
        raise ValueError('mutinfo must be a 2D matrix')

    kwargs.setdefault('interpolation', 'nearest')
    kwargs.setdefault('origin', 'lower')
    kwargs.setdefault('aspect', 'auto')

    xlabel = kwargs.pop('xlabel', None)
    ylabel = kwargs.pop('ylabel', None)
    title = kwargs.pop('title', None)

    import matplotlib.pyplot as plt
    show = plt.imshow(heatmap, *args, **kwargs), plt.colorbar()

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if title is not None:
        plt.title(title)
    return show


def parseHeatmap(heatmap, **kwargs):
    """Returns a two dimensional array and a dictionary with information parsed
    from *heatmap*, which may be an input stream or an :file:`.hm` file in VMD
    plugin Heat Mapper format."""

    try:
        readline, close = heatmap.readline, lambda: None
    except AttributeError:
        heatmap = openFile(heatmap)
        readline, close = heatmap.readline, heatmap.close

    meta = {}
    arrs = []

    line = readline()
    while line:
        if line.startswith('-'):
            label, data = line[1:].split(None, 1)
            data = data.strip()
            if data[0] == data[-1] == '"':
                data = data[1:-1]
            label = label.strip()
            try:
                meta[label] = HMTYPES[label](data)
            except KeyError:
                LOGGER.warn('Unrecognized label encountered: {0}'
                            .format(repr(label)))
                meta[label] = HMTYPES[label](data)
            except TypeError:
                LOGGER.warn('Could not parse data with label {0}.'
                            .format(repr(label)))
        else:
            arrs.append(line.rstrip())
        line = readline()
    close()
    nnums = len(meta.get('numbering', ''))
    heatmap = []
    numbers = []

    for arr in arrs:
        if nnums:
            items = arr.split(':', nnums + 1)
            numbers.append(items[:nnums])
        else:
            items = [arr]

        try:
            heatmap.append(fromstring(items[-1], float, sep=';'))
        except:
            heatmap.append(frombuffer(items[-1], float, sep=';'))

    heatmap = array(heatmap)
    if nnums:
        numbering = meta['numbering']
        try:
            numbers = array(numbers, int)
        except ValueError:
            try:
                numbers = array(numbers, float)
            except ValueError:
                LOGGER.warn('Numbering for y-axis could not be parsed.')
                numbering = []
        for i, label in enumerate(numbering):
            meta[label] = numbers[:, i].copy()

    return heatmap, meta


def writeHeatmap(filename, heatmap, **kwargs):
    """Returns *filename* that contains *heatmap* in Heat Mapper :file:`.hm`
    file (extension is automatically added when not found).  *filename* may
    also be an output stream.

    :arg title: title of the heatmap
    :type title: str

    :arg xlabel: x-axis lab, default is ``'unknown'``
    :type xlabel: str

    :arg ylabel: y-axis lab, default is ``'unknown'``
    :type ylabel: str

    :arg xorigin: x-axis origin, default is 0
    :type xorigin: float

    :arg xstep: x-axis step, default is 1
    :type xstep: float

    :arg min: minimum value, default is minimum in *heatmap*
    :type min: float

    :arg max: maximum value, default is maximum in *heatmap*
    :type max: float

    :arg format: number format, default is ``'%f'``
    :type format: str

    Other keyword arguments that are arrays with length equal to the y-axis
    (second dimension of heatmap) will be considered as *numbering*."""

    try:
        ndim, shape = heatmap.ndim, heatmap.shape
    except:
        raise TypeError('heatmap must be an array object')
    if ndim!= 2:
        raise TypeError('heatmap must be a 2D array')

    try:
        write, close, stream = filename.write, lambda: None, filename
    except AttributeError:
        out = openFile(addext(filename, '.hm'), 'w')
        write, close, stream = out.write, out.close, out

    format = kwargs.pop('format', '%f')
    write('-min "{0}"\n'.format(kwargs.pop('min', heatmap.min())))
    write('-max "{0}"\n'.format(kwargs.pop('max', heatmap.max())))
    for label, default in [
        ('title', 'unknown'),
        ('xlabel', 'unknown'),
        ('xorigin', 0),
        ('xstep', 1),
        ('ylabel', 'unknown'),
    ]:
        write('-{0} "{1}"\n'.format(label, kwargs.pop(label, default)))

    numbering = []
    numlabels = []
    for key, val in kwargs.items():
        try:
            length = len(val)
        except TypeError:
            LOGGER.warn('Keyword argument {0} is not used.'.format(key))
        else:
            if length == shape[0]:
                numlabels.append(key)
                numbering.append(val)
    if not numbering:
        numlabels.append('unknown')
        numbering.append(arange(1, shape[0] + 1))

    write('-numbering "{0}"\n'.format(':'.join(numlabels)))

    for i, row in enumerate(heatmap):
        write(':'.join(str(nums[i]) for nums in numbering) + ':')
        row.tofile(stream, sep=';', format=format)
        write(';\n')

    close()
    return filename



# -*- coding: utf-8 -*-
"""This module provides :class:`.KDTree` class as an interface to Thomas
Hamelryck's KDTree C module distributed with Biopython."""

from .kdtree import KDTree

__all__ = ['KDTree']



"""This module contains unit tests for :mod:`~prody.proteins`."""

import numpy as np
from numpy.testing import *
from prody.utilities import importDec
dec = importDec()

from prody import *
from prody import LOGGER
from prody.tests import unittest
from prody.tests.datafiles import *

import os

LOGGER.verbosity = 'none'

PATH_PDB = pathDatafile('3enl_addH')
SPLIT_PATH = os.path.split(PATH_PDB)
TARGET_PATH = os.path.join(SPLIT_PATH[0], 'wb_cluster_' + SPLIT_PATH[1])

class TestWaterBridges(unittest.TestCase):

    def setUp(self):
        """Parse the PDB file and run the core calculations."""
        self.atoms = parsePDB(PATH_PDB)

        if prody.PY3K:
            self.waterBridgesChain = calcWaterBridges(self.atoms)
            self.chainMultiWater = self.waterBridgesChain[37]

            self.waterBridgesClust = calcWaterBridges(self.atoms, method='cluster')
            self.clustMultiWater = self.waterBridgesClust[36]

            self.waterBridgesClustInfo = calcWaterBridges(self.atoms, method='cluster', 
                                                        output='info')
            self.clustInfoMultiWater = self.waterBridgesClustInfo[36]

            self.pro0 = self.clustMultiWater.proteins[0]
            self.pro1 = self.clustMultiWater.proteins[1]
            self.w0 = self.clustMultiWater.waters[0]
            self.w1 = self.clustMultiWater.waters[1]
        
    @dec.slow
    def testSingleWaterBridgesChain(self):
        if prody.PY3K:
            self.assertIsInstance(self.waterBridgesChain, list,
                                  'calcWaterBridges chain output should be a list')
            
            self.assertEqual(len(self.waterBridgesChain), 77,
                             'calcWaterBridges chain list should have 77 items')
            
            
            self.assertIsInstance(self.chainMultiWater, prody.proteins.waterbridges.AtomicOutput,
                                  'calcWaterBridges output should contain AtomicOutput')
            
            self.assertEqual(len(self.chainMultiWater.proteins), 2,
                             'water bridges chains should have 2 protein items')
            self.assertIsInstance(self.chainMultiWater.proteins[0], Atom,
                                  'protein items in water bridges chains should be Atom')
            
            self.assertEqual(len(self.chainMultiWater.waters), 2,
                             'water bridges chain should have 2 water items')
            self.assertIsInstance(self.chainMultiWater.waters[0], Atom,
                                  'waters items in water bridges chains should be Atom')
            
    @dec.slow
    def testSingleWaterBridgesCluster(self):
        if prody.PY3K:
            self.assertIsInstance(self.waterBridgesClust, list,
                                  'calcWaterBridges clust output should be a list')
            
            self.assertEqual(len(self.waterBridgesClust), 74,
                             'calcWaterBridges clust list should have 74 items')
            
            self.assertIsInstance(self.clustMultiWater, prody.proteins.waterbridges.AtomicOutput,
                                  'calcWaterBridges output should contain AtomicOutput')
            
            self.assertEqual(len(self.clustMultiWater.proteins), 2,
                             'tested water bridges multi-water cluster should have 3 protein items')
            self.assertIsInstance(self.clustMultiWater.proteins[0], Atom,
                                  'protein items in water bridges clusters should be Atom')
            
            self.assertEqual(len(self.clustMultiWater.waters), 2,
                             'tested water bridges multi-water cluster should have 2 waters items')
            self.assertIsInstance(self.clustMultiWater.waters[0], Atom,
                                  'waters items in water bridges clusters should be Atom')

    @dec.slow
    def testSaveWaterBridges(self):
        if prody.PY3K:
            savePDBWaterBridges(self.waterBridgesClust, self.atoms, TARGET_PATH)
            self.saved = parsePDB(TARGET_PATH)

            self.assertFalse(np.allclose(self.atoms.protein.getBetas(), self.saved.protein.getBetas()),
                             'savePDBWaterBridges did not change protein b-factors')
            
            self.assertFalse(np.allclose(self.atoms.protein.getOccupancies(), self.saved.protein.getOccupancies()),
                             'savePDBWaterBridges did not change protein occupancies')
            
            self.assertFalse(np.equal(self.atoms.water.numAtoms(), self.saved.water.numAtoms()),
                             'savePDBWaterBridges did not change the number of water atoms')
            
    @dec.slow
    def testSingleWaterBridgesOutputInfo(self):
        if prody.PY3K:
            self.assertIsInstance(self.waterBridgesClustInfo, list,
                                  'calcWaterBridges clust info output should be a list')
            
            self.assertEqual(len(self.waterBridgesClustInfo), 74,
                             'calcWaterBridges clust info list should have 74 items')
            
            self.assertIsInstance(self.clustInfoMultiWater, list,
                                  'tested water bridges multi-water cluster info output should contain lists')
            
            self.assertEqual(len(self.clustInfoMultiWater), 9,
                             'tested water bridges multi-water cluster info should have 9 items')
            self.assertIsInstance(self.clustInfoMultiWater[0], str,
                                  'info items in water bridges clusters should be str')


            self.assertEqual(self.clustInfoMultiWater[0], self.pro0.getResname() + str(self.pro0.getResnum()),
                             'item 0 from selected info should be resname + resnum of atom 0')
            self.assertEqual(self.clustInfoMultiWater[1], self.pro0.getName() + '_' + str(self.pro0.getIndex()),
                             'item 1 from selected info should be name + _ + index of atom 0')
            self.assertEqual(self.clustInfoMultiWater[2], self.pro0.getChid(),
                             'item 2 from selected info should be chid of atom 0')

            self.assertEqual(self.clustInfoMultiWater[3], self.pro1.getResname() + str(self.pro1.getResnum()),
                             'item 3 from selected info should be resname + resnum of atom 1')
            self.assertEqual(self.clustInfoMultiWater[4], self.pro1.getName() + '_' + str(self.pro1.getIndex()),
                             'item 4 from selected info should be name + _ + index of atom 1')
            self.assertEqual(self.clustInfoMultiWater[5], self.pro1.getChid(),
                             'item 5 from selected info should be chid of atom 1')

            self.assertEqual(self.clustInfoMultiWater[6], calcDistance(self.pro0, self.pro1),
                             'item 6 from selected info should be distance from protein atom 0 to protein atom 1')
            self.assertEqual(self.clustInfoMultiWater[7], len(self.clustMultiWater.waters),
                             'item 7 from selected info should be number of waters in the cluster')
            self.assertEqual(self.clustInfoMultiWater[8], 
                             list(map(lambda w: w.getChid() + "_" + str(w.getIndex()), self.clustMultiWater.waters)),
                             'item 8 from selected info should be chid and index for 2 water atoms')
           



"""This module contains unit tests for :mod:`~prody.KDTree` module."""

from numpy import tile, array, arange, ones
from numpy.testing import assert_allclose

from prody.tests import unittest
from prody.kdtree import KDTree
ATOL = 1e-5
RTOL = 0

class TestKDTree(unittest.TestCase):


    def setUp(self):

        self.coords = tile(arange(10), (3,1)).T.astype(float)
        self.kdtree = KDTree(self.coords)

    def testSearch(self):

        kdtree = self.kdtree
        coords = self.coords
        kdtree.search(10000, array([0,0,0]))
        indices = kdtree.getIndices()
        dist = (coords**2).sum(1)**0.5
        radii = kdtree.getDistances()
        assert_allclose(radii, dist[indices], rtol=RTOL, atol=ATOL,
                        err_msg='KDTree search failed')

    def testAllSearch(self):
        kdtree = self.kdtree
        coords = self.coords
        kdtree.search(1.75)
        radii = kdtree.getDistances()
        indices = kdtree.getIndices()
        self.assertEqual(len(indices), 9, 'KDTree all search failed')
        for pair, radius in zip(indices, radii):
            x, y = coords[pair]
            assert_allclose(((x - y)**2).sum()**0.5, radius,
                            rtol=RTOL, atol=ATOL,
                            err_msg='KDTree all search failed')


COORDS = array([[-1., -1., 0.],
                [-1.,  5., 0.],
                [ 2.,  2., 0.],
                [ 5., -1., 0.],
                [ 5.,  5., 0.],])
UNITCELL = array([4., 4., 0.])

KDTREE = KDTree(COORDS)
KDTREE_PBC = KDTree(COORDS, unitcell=UNITCELL)

class TestKDTreePBC(unittest.TestCase):

    def testPointNoPBC(self):

        KDTREE.search(2, ones(3) * 2)
        self.assertEqual(1, KDTREE.getCount())

    def testPairNoPBC(self):

        KDTREE.search(2)
        self.assertEqual(0, KDTREE.getCount())

    def testPointPBC(self):

        KDTREE_PBC.search(2, array([2., 2., 0.]))
        self.assertEqual(5, KDTREE_PBC.getCount())

    def testPairPBC(self):

        KDTREE_PBC.search(2)
        self.assertEqual(8, KDTREE_PBC.getCount())

